# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import pytest
from affine import Affine

from odc.geo import CRS, geom, ixy_, resxy_, resyx_, wh_, xy_
from odc.geo._interop import have
from odc.geo.geobox import (
    AnchorEnum,
    GeoBox,
    bounding_box_in_pixel_domain,
    gbox_boundary,
    geobox_intersection_conservative,
    geobox_union_conservative,
    scaled_down_geobox,
)
from odc.geo.math import apply_affine, is_affine_st
from odc.geo.testutils import (
    epsg3577,
    epsg3857,
    epsg4326,
    esri54019,
    mkA,
    modis_crs,
    xy_from_gbox,
    xy_norm,
)

# pylint: disable=pointless-statement,too-many-statements,protected-access


def test_geobox_simple():
    t = GeoBox(
        (4000, 4000), Affine(0.00025, 0.0, 151.0, 0.0, -0.00025, -29.0), epsg4326
    )

    expect_lon = np.asarray(
        [
            151.000125,
            151.000375,
            151.000625,
            151.000875,
            151.001125,
            151.001375,
            151.001625,
            151.001875,
            151.002125,
            151.002375,
        ]
    )

    expect_lat = np.asarray(
        [
            -29.000125,
            -29.000375,
            -29.000625,
            -29.000875,
            -29.001125,
            -29.001375,
            -29.001625,
            -29.001875,
            -29.002125,
            -29.002375,
        ]
    )
    expect_resolution = pytest.approx((-0.00025, 0.00025))
    assert t.resolution.yx == expect_resolution
    assert t.rotate(33).resolution.yx == expect_resolution
    assert t.rotate(-11.37).resolution.yx == expect_resolution

    assert t.coordinates["latitude"].values.shape == (4000,)
    assert t.coordinates["longitude"].values.shape == (4000,)

    np.testing.assert_almost_equal(t.coords["latitude"].values[:10], expect_lat)
    np.testing.assert_almost_equal(t.coords["longitude"].values[:10], expect_lon)

    assert (t == "some random thing") is False

    # ensure GeoBox accepts string CRS
    assert isinstance(
        GeoBox(
            (4000, 4000), Affine(0.00025, 0.0, 151.0, 0.0, -0.00025, -29.0), "epsg:4326"
        ).crs,
        CRS,
    )

    # Check GeoBox class is hashable
    t_copy = GeoBox(t.shape, t.transform, t.crs)
    t_other = GeoBox(wh_(t.width + 1, t.height), t.transform, t.crs)
    assert t_copy is not t
    assert t == t_copy
    assert len({t, t, t_copy}) == 1
    assert len({t, t_copy, t_other}) == 2


def test_xy_from_geobox():
    gbox = GeoBox(wh_(3, 7), Affine.translation(10, 1000), epsg3857)
    xx, yy = xy_from_gbox(gbox)

    assert xx.shape == gbox.shape
    assert yy.shape == gbox.shape
    assert (xx[:, 0] == 10.5).all()
    assert (xx[:, 1] == 11.5).all()
    assert (yy[0, :] == 1000.5).all()
    assert (yy[6, :] == 1006.5).all()

    xx_, yy_, A = xy_norm(xx, yy)
    assert xx_.shape == xx.shape
    assert yy_.shape == yy.shape
    np.testing.assert_almost_equal((xx_.min(), xx_.max()), (0, 1))
    np.testing.assert_almost_equal((yy_.min(), yy_.max()), (0, 1))
    assert (xx_[0] - xx_[1]).sum() != 0
    assert (xx_[:, 0] - xx_[:, 1]).sum() != 0

    XX, YY = apply_affine(A, xx_, yy_)
    np.testing.assert_array_almost_equal(xx, XX)
    np.testing.assert_array_almost_equal(yy, YY)


def test_geobox():
    points_list = [
        [
            (148.2697, -35.20111),
            (149.31254, -35.20111),
            (149.31254, -36.331431),
            (148.2697, -36.331431),
        ],
        [
            (148.2697, 35.20111),
            (149.31254, 35.20111),
            (149.31254, 36.331431),
            (148.2697, 36.331431),
        ],
        [
            (-148.2697, 35.20111),
            (-149.31254, 35.20111),
            (-149.31254, 36.331431),
            (-148.2697, 36.331431),
        ],
        [
            (-148.2697, -35.20111),
            (-149.31254, -35.20111),
            (-149.31254, -36.331431),
            (-148.2697, -36.331431),
            (148.2697, -35.20111),
        ],
    ]
    for points in points_list:
        polygon = geom.polygon(points, crs=epsg3577)
        resolution = resyx_(-25, 25)
        geobox = GeoBox.from_geopolygon(polygon, resolution)

        # check single value resolution equivalence
        assert GeoBox.from_geopolygon(polygon, 25) == geobox
        assert GeoBox.from_geopolygon(polygon, 25.0) == geobox

        assert GeoBox.from_geopolygon(polygon, resolution, crs=geobox.crs) == geobox

        # check that extra padding added by alignment is smaller than pixel size
        assert abs(resolution.x) > abs(
            geobox.extent.boundingbox.left - polygon.boundingbox.left
        )
        assert abs(resolution.x) > abs(
            geobox.extent.boundingbox.right - polygon.boundingbox.right
        )
        assert abs(resolution.y) > abs(
            geobox.extent.boundingbox.top - polygon.boundingbox.top
        )
        assert abs(resolution.y) > abs(
            geobox.extent.boundingbox.bottom - polygon.boundingbox.bottom
        )

    A = mkA(0, scale=(10, -10), translation=(-48800, -2983006))

    w, h = 512, 256
    gbox = GeoBox(wh_(w, h), A, epsg3577)
    assert gbox.boundingbox.crs is epsg3577
    assert gbox.footprint("epsg:4326") == gbox.geographic_extent

    assert gbox.shape == (h, w)
    assert gbox.transform == A
    assert gbox.extent.crs == gbox.crs
    assert gbox.geographic_extent.crs == epsg4326
    assert gbox.extent.boundingbox.height == h * 10.0
    assert gbox.extent.boundingbox.width == w * 10.0
    assert gbox.alignment.yx == (4, 0)  # 4 because -2983006 % 10 is 4
    assert isinstance(str(gbox), str)
    assert "EPSG:3577" in repr(gbox)
    assert gbox.axis_aligned is True

    assert GeoBox((1, 1), mkA(0), epsg4326).geographic_extent.crs == epsg4326
    assert GeoBox((1, 1), mkA(0), None).dimensions == ("y", "x")

    g2 = gbox[:-10, :-20]
    assert g2.shape == (gbox.height - 10, gbox.width - 20)

    # step of 1 is ok
    g2 = gbox[::1, ::1]
    assert g2.shape == gbox.shape

    assert gbox[0].shape == (1, gbox.width)
    assert gbox[:3].shape == (3, gbox.width)
    assert gbox[0, 0] == gbox[:1, :1]
    assert gbox[0, 0:1] == gbox[:1, :1]
    assert gbox[0:1, 1] == gbox[:1, 1:2]

    assert gbox[gbox.extent] == gbox
    assert gbox[gbox.geographic_extent] == gbox
    assert gbox[gbox] == gbox
    assert gbox[gbox[3:6, :4].extent.boundingbox] == gbox[3:6, :4]
    assert gbox[gbox.extent.centroid].shape.wh == (1, 1)
    assert gbox[geom.point(0.5, 0.5, None)] == gbox[:1, :1]
    assert gbox[geom.point(0, 0.9, None)] == gbox[:1, :1]

    with pytest.raises(NotImplementedError):
        gbox[::2, :]

    # too many slices
    with pytest.raises(ValueError):
        gbox[:1, :1, :]

    assert gbox.buffered(0, 10).shape == (gbox.height + 2 * 1, gbox.width)
    assert gbox.buffered(10).shape == (gbox.height + 2 * 1, gbox.width + 2 * 1)
    assert gbox.buffered(20, 30).shape == (gbox.height + 2 * 3, gbox.width + 2 * 2)

    assert (gbox | gbox) == gbox
    assert (gbox & gbox) == gbox
    assert gbox.is_empty() is False
    assert bool(gbox) is True

    assert (gbox[:3, :4] & gbox[3:, 4:]).is_empty()
    assert (gbox[:3, :4] & gbox[:3, 4:]).is_empty()
    assert (gbox[:3, :4] & gbox[30:, 40:]).is_empty()

    assert gbox[:3, :7].center_pixel == gbox[1, 3]
    assert gbox[:2, :4].center_pixel == gbox[1, 2]

    with pytest.raises(ValueError):
        geobox_intersection_conservative([])

    with pytest.raises(ValueError):
        geobox_union_conservative([])

    # can not combine across CRSs
    with pytest.raises(ValueError):
        bounding_box_in_pixel_domain(
            GeoBox((1, 1), mkA(0), epsg4326), GeoBox(ixy_(2, 3), mkA(0), epsg3577)
        )


def test_gbox_overlap_roi():
    gbox = GeoBox(wh_(100, 30), Affine.translation(0, 0), epsg4326)
    assert gbox.overlap_roi(gbox[3:7, 13:29]) == np.s_[3:7, 13:29]
    assert gbox[10:20, 5:10].overlap_roi(gbox) == np.s_[0:10, 0:5]

    a = gbox[10:20, 20:40]
    b = gbox[11:30, 10:25]
    assert a.overlap_roi(b) == np.s_[1:10, 0:5]


def test_gbox_boundary():
    geobox = GeoBox(wh_(6, 2), Affine.translation(0, 0), epsg4326)

    bb = gbox_boundary(geobox, 3)

    assert bb.shape == (4 + (3 - 2) * 4, 2)
    assert set(bb.T[0]) == {0.0, 3.0, 6.0}
    assert set(bb.T[1]) == {0.0, 1.0, 2.0}

    assert geobox.map_bounds() == geobox.boundingbox.map_bounds()

    geobox = GeoBox(wh_(6, 2), Affine.translation(0, 0), None)
    assert geobox.boundingbox.crs is None
    assert geobox.map_bounds() == geobox.boundingbox.map_bounds()


def test_geobox_scale_down():
    crs = CRS("EPSG:3857")

    A = mkA(0, (111.2, 111.2), translation=(125671, 251465))
    for s in [2, 3, 4, 8, 13, 16]:
        gbox = GeoBox(wh_(233 * s, 755 * s), A, crs)
        gbox_ = scaled_down_geobox(gbox, s)

        assert gbox_.width == 233
        assert gbox_.height == 755
        assert gbox_.crs is crs
        assert gbox_.extent.contains(gbox.extent)
        assert gbox.extent.difference(gbox.extent).area == 0.0

    gbox = GeoBox((1, 1), A, crs)
    for s in [2, 3, 5]:
        gbox_ = scaled_down_geobox(gbox, 3)

        assert gbox_.shape == (1, 1)
        assert gbox_.crs is crs
        assert gbox_.extent.contains(gbox.extent)


@pytest.mark.parametrize(
    "geobox",
    [
        GeoBox.from_bbox((-10, -2, 5, 4), "epsg:4326", tight=True, resolution=0.2),
        GeoBox.from_bbox((-10, -2, 5, 4), "epsg:3857", tight=True, resolution=1),
        GeoBox.from_bbox(
            (-10, -2, 5, 4), "epsg:3857", tight=True, resolution=resxy_(1, 2)
        ),
    ],
)
@pytest.mark.parametrize("shape", [256, (128, 128), (33, 11)])
def test_zoom_to_shape(geobox: GeoBox, shape):
    assert geobox.zoom_to(shape).crs == geobox.crs

    if isinstance(shape, int):
        assert max(geobox.zoom_to(shape).shape) == shape
    else:
        assert geobox.zoom_to(shape).shape == shape


def test_zoom_to_resolution():
    geobox = GeoBox.from_bbox((-3, -4, 3, 4), epsg4326, resolution=1)
    assert geobox.shape == (8, 6) and geobox.resolution.xy == (1, -1)
    assert geobox.zoom_to(resolution=2).resolution.xy == (2, -2)
    assert geobox.zoom_to(resolution=2).zoom_to(resolution=1) == geobox

    with pytest.raises(ValueError):
        geobox.zoom_to()


def test_non_st():
    A = mkA(rot=10, translation=(-10, 20), shear=0.3)
    assert is_affine_st(A) is False

    gbox = GeoBox((1, 1), A, "epsg:3857")

    assert gbox._confirm_axis_aligned() is False
    assert gbox.axis_aligned is False

    with pytest.raises(ValueError):
        assert gbox.coordinates


def test_from_polygon():
    box = geom.box(1, 13, 17, 37, "epsg:4326")
    gbox = GeoBox.from_geopolygon(box, 0.1)
    assert gbox.crs == box.crs
    assert gbox.extent.boundingbox == box.boundingbox
    assert gbox.resolution.y < 0

    # check that align= still works by name and by order
    assert gbox == GeoBox.from_geopolygon(gbox.extent, gbox.resolution, None, xy_(0, 0))
    assert gbox == GeoBox.from_geopolygon(gbox.extent, gbox.resolution, align=xy_(0, 0))

    box = geom.box(-1, 10, 2, 12, "epsg:4326")
    gbox = GeoBox.from_geopolygon(box, tight=True, shape=300)
    assert gbox.crs == box.crs
    assert gbox.shape.wh == (300, 200)
    assert gbox.resolution.x == -gbox.resolution.y
    assert gbox.aspect == 300 / 200

    gbox = GeoBox.from_geopolygon(box, tight=True, shape=(100, 29))
    assert gbox.crs == box.crs
    assert gbox.shape.wh == (29, 100)

    gbox = GeoBox.from_geopolygon(box, crs="utm", tight=True, shape=(100, 29))
    assert gbox.crs != box.crs
    assert gbox.shape.wh == (29, 100)
    assert 32601 <= gbox.crs.epsg <= 32660


def test_from_polygon_compat_align():
    box = geom.box(1, 13, 17, 37, "epsg:4326")
    assert GeoBox.from_geopolygon(box, 2, align=xy_(1, 1)) == GeoBox.from_geopolygon(
        box, 2, anchor=xy_(0.5, 0.5)
    )


def test_from_bbox():
    bbox = (1, 13, 17, 37)
    shape = (23, 47)
    gbox = GeoBox.from_bbox(bbox, tight=True, shape=shape)
    assert gbox.shape == shape
    assert gbox.crs == "epsg:4326"
    assert gbox.extent.boundingbox == bbox
    assert gbox.resolution.y < 0

    assert (
        GeoBox.from_bbox(bbox, gbox.crs, tight=True, resolution=gbox.resolution) == gbox
    )

    assert GeoBox.from_bbox(bbox, shape=shape, tight=False) != gbox
    assert GeoBox.from_bbox(bbox, resolution=gbox.resolution, tight=False) != gbox

    assert GeoBox.from_bbox(geom.BoundingBox(*bbox), shape=shape).crs == "epsg:4326"
    assert (
        GeoBox.from_bbox(geom.BoundingBox(*bbox, "epsg:3857"), shape=shape).crs
        == "epsg:3857"
    )

    gbox = GeoBox.from_bbox((0, 0, 2, 3), tight=True, shape=300, crs="epsg:3857")
    assert gbox.shape.wh == (200, 300)
    assert gbox.resolution.x == -gbox.resolution.y

    gbox = GeoBox.from_bbox((-1, 0, 2, 2), tight=True, shape=300, crs="epsg:3857")
    assert gbox.shape.wh == (300, 200)
    assert gbox.resolution.x == -gbox.resolution.y

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.FLOATING
    ) == GeoBox.from_bbox(bbox, shape=shape, tight=True)

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.FLOATING
    ) == GeoBox.from_bbox(bbox, shape=shape, anchor="floating")

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.EDGE
    ) == GeoBox.from_bbox(bbox, shape=shape, anchor=xy_(0, 0))

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.EDGE
    ) != GeoBox.from_bbox(bbox, shape=shape, anchor=xy_(0.3, 0))

    assert GeoBox.from_bbox(bbox, shape=shape, anchor=0.3) == GeoBox.from_bbox(
        bbox, shape=shape, anchor=xy_(0.3, 0.3)
    )

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.EDGE
    ) == GeoBox.from_bbox(bbox, shape=shape, anchor=0)

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.EDGE
    ) == GeoBox.from_bbox(bbox, shape=shape)

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.CENTER
    ) == GeoBox.from_bbox(bbox, shape=shape, anchor=xy_(0.5, 0.5))

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.CENTER
    ) == GeoBox.from_bbox(bbox, shape=shape, anchor=0.5)

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.CENTER
    ) == GeoBox.from_bbox(bbox, shape=shape, anchor="center")

    assert GeoBox.from_bbox(
        bbox, shape=shape, anchor=AnchorEnum.CENTER
    ) == GeoBox.from_bbox(bbox, shape=shape, anchor="centre")

    assert GeoBox.from_bbox((1, 40, 2, 43), "utm", shape=(12, 12)).crs == "epsg:32631"

    # one of resolution= or shape= must be supplied
    with pytest.raises(ValueError):
        _ = GeoBox.from_bbox(bbox)


def test_outline():
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))
    assert gbox.outline() == gbox.outline("native")
    assert gbox.outline().crs == gbox.crs
    assert gbox.outline("geo").crs == "epsg:4326"
    assert gbox.outline("pixel").boundingbox == (0, 0, 200, 100)
    assert gbox.outline("pixel").crs is None
    assert gbox.outline(notch=0).geom_type == "GeometryCollection"


def test_footprint():
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))
    assert gbox.footprint("epsg:4326").crs == "epsg:4326"
    assert gbox.footprint("utm").crs.proj.utm_zone is not None
    assert gbox.footprint("utm-s").crs.proj.utm_zone.endswith("S")


def test_to_crs():
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))
    assert gbox.to_crs("utm").crs == "epsg:32631"
    assert gbox.to_crs("utm") == gbox.to_crs("epsg:32631")
    assert gbox.to_crs("utm").extent.contains(gbox.extent.to_crs("epsg:32631"))
    assert gbox.to_crs("epsg:4326").extent.contains(gbox.geographic_extent)
    assert gbox.to_crs("utm-n") == gbox.to_crs("utm")
    assert gbox.to_crs("utm-s").crs.proj.utm_zone == "31S"

    gbox = GeoBox.from_bbox([0, -50, 20, -10], "epsg:3857", shape=wh_(200, 100))
    assert gbox.to_crs("utm").crs.proj.utm_zone == "31S"
    assert gbox.to_crs("utm-n").crs.proj.utm_zone == "31N"


def test_snap_to():
    def aligned(a, b):
        try:
            _ = a & b
        except ValueError:
            return False
        return True

    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))
    assert aligned(gbox, gbox[2:, 3:])
    assert gbox.snap_to(gbox[2:, 3:]) == gbox

    gbox_ = gbox.center_pixel.translate_pix(0.1, 0.2).pad(10)
    assert not aligned(gbox, gbox_)
    assert aligned(gbox, gbox_.snap_to(gbox))
    assert aligned(gbox.snap_to(gbox_), gbox_)

    assert gbox.snap_to(gbox_) != gbox
    assert gbox.snap_to(gbox_).shape == gbox.shape
    assert gbox.snap_to(gbox_).crs == gbox.crs


def test_svg():
    # smoke test only
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))
    assert isinstance(gbox.svg(), str)
    assert len(gbox.svg()) > 0
    assert gbox.svg(10) != gbox.svg(1)

    assert gbox.grid_lines(mode="pixel").crs == "epsg:3857"

    assert isinstance(gbox._repr_svg_(), str)

    # empty should still work
    assert isinstance((gbox[:3] & gbox[3:])._repr_svg_(), str)


def test_html_repr():
    # smoke test only
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))
    assert isinstance(gbox._repr_html_(), str)
    assert ">EPSG<" in gbox._repr_html_()
    # empties should still work
    assert isinstance(gbox[:0]._repr_html_(), str)
    assert isinstance(gbox[:0, :0]._repr_html_(), str)

    # non-epsg authority
    gbox = GeoBox.from_bbox((0, 0, 100, 200), esri54019, resolution=1)
    assert ">ESRI<" in gbox._repr_html_()

    # no authority CRS
    assert ">EPSG<" not in gbox.to_crs(modis_crs)._repr_html_()

    # no crs case
    gbox = GeoBox(wh_(200, 100), Affine.translation(-10, 20), None)
    assert gbox.crs is None
    assert isinstance(gbox._repr_html_(), str)
    assert ">EPSG<" not in gbox._repr_html_()

    # empty should still work
    assert isinstance((gbox[:3] & gbox[3:])._repr_html_(), str)

    # empty should still work
    assert isinstance((gbox[:0, :0])._repr_html_(), str)


def test_lrtb():
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))

    assert gbox.right.left == gbox
    assert gbox.left.right == gbox
    assert gbox.top.bottom == gbox
    assert gbox.bottom.top == gbox

    assert gbox.left == GeoBox.from_bbox([-20, 0, 0, 10], gbox.crs, shape=gbox.shape)
    assert gbox.right == GeoBox.from_bbox([20, 0, 40, 10], gbox.crs, shape=gbox.shape)

    assert gbox.bottom == GeoBox.from_bbox([0, -10, 20, 0], gbox.crs, shape=gbox.shape)
    assert gbox.top == GeoBox.from_bbox([0, 10, 20, 20], gbox.crs, shape=gbox.shape)


def test_compat():
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))
    gbox_ = gbox.compat
    if gbox_ is None:  # no datacube in this environment
        return
    assert gbox.width == gbox_.width
    assert gbox.height == gbox_.height
    assert gbox.affine == gbox_.affine
    assert gbox.crs == str(gbox.crs)


def test_project():
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))

    pix = gbox.outline("pixel", notch=0)
    assert pix.crs is None
    wld = gbox.project(pix)
    assert wld.crs == gbox.crs
    assert wld == gbox.outline("native", notch=0)

    assert gbox.project(wld).crs is None
    assert gbox.project(wld.to_crs("epsg:4326")).crs is None

    assert gbox.project(wld).buffer(0.001).contains(pix)
    assert gbox.project(wld.to_crs("epsg:4326")).buffer(0.001).contains(pix)


def test_enclosing():
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))

    assert (gbox.enclosing(gbox.center_pixel.geographic_extent)) | gbox == gbox
    assert (
        gbox.enclosing(gbox.center_pixel.geographic_extent.boundingbox)
    ) | gbox == gbox

    gbox2 = gbox.bottom.left
    region = gbox2.extent.buffer(-0.01)
    assert gbox.enclosing(region) == gbox2
    assert gbox.enclosing(region.boundingbox) == gbox2
    assert gbox.enclosing(region.to_crs("utm")) == gbox2
    assert (gbox.enclosing(region) & gbox).is_empty
    assert (gbox.enclosing(region) | gbox) == gbox | gbox2

    assert gbox.enclosing(gbox.extent.centroid).shape.wh == (1, 1)

    with pytest.raises(ValueError):
        _ = gbox.enclosing(geom.point(0, 0, None))


@pytest.mark.parametrize("n", [10, 100, 23])
@pytest.mark.parametrize("with_edges", [False, True])
def test_qr2sample(n, with_edges):
    gbox = GeoBox.from_bbox([0, 0, 20, 10], "epsg:3857", shape=wh_(200, 100))
    xx1 = gbox.qr2sample(n, with_edges=with_edges)
    xx1_ = gbox.qr2sample(n, with_edges=with_edges)
    xx2 = gbox.qr2sample(n, with_edges=with_edges, offset=1000)
    assert xx1.crs is None
    assert (xx1 ^ xx1_).is_empty
    assert not (xx1 ^ xx2).is_empty
    assert (gbox.project(xx1) - gbox.extent).is_empty
    assert (xx1 - gbox.project(gbox.extent)).is_empty
    assert (xx2 - gbox.project(gbox.extent)).is_empty


_gbox = GeoBox((3, 17), mkA(scale=(10, -10), translation=(1000, -20)), epsg3857)


@pytest.mark.parametrize(
    ("gbox", "shape"),
    [
        (_gbox, (1, 1)),
        (_gbox, (100, 200)),
        (_gbox, (200, 100)),
    ],
)
def test_crop(gbox, shape):
    assert gbox.crop(shape).shape == shape
    assert gbox.crop(shape).crs == gbox.crs
    assert gbox.crop(shape)[:1, :1] == gbox[:1, :1]
    assert gbox.crop(shape) == gbox.expand(shape)


@pytest.mark.skipif(have.folium is False, reason="No folium installed")
@pytest.mark.parametrize(
    "geobox",
    [
        GeoBox.from_bbox((-10, -2, 5, 4), "epsg:4326", tight=True, resolution=0.2),
        GeoBox.from_bbox((-10, -2, 5, 4), "epsg:3857", tight=True, resolution=1),
        GeoBox.from_bbox(
            (-10, -2, 5, 4), "epsg:3857", tight=True, resolution=resxy_(1, 2)
        ),
    ],
)
def test_explore_geobox(geobox):
    from folium import GeoJson, Map

    # Test explore on dataset input and verify that output is a folium map
    # that contains a GeoJson layer
    m = geobox.explore()
    assert isinstance(m, Map)
    assert sum(isinstance(child, GeoJson) for child in m._children.values()) == 2

    # Test that we can turn off gridlines
    m = geobox.explore(grid_lines=False)
    assert sum(isinstance(child, GeoJson) for child in m._children.values()) == 1

    # Verify that passing a custom map works correctly
    m_external = Map()
    geobox.explore(map=m_external)
    assert isinstance(m_external, Map)
    assert (
        sum(isinstance(child, GeoJson) for child in m_external._children.values()) == 2
    )


@pytest.mark.parametrize(
    ("shape", "allowed"),
    [
        pytest.param((100, 100), True, id="positive xy: normal"),
        pytest.param((0, 100), False, id="zero x: can't calculate resolution"),
        pytest.param((100, 0), False, id="zero y: can't calculate resolution"),
        pytest.param((-100, 100), False, id="negative x: not allowed"),
        pytest.param((100, -100), False, id="negative y: not allowed"),
        pytest.param((-100, -100), False, id="both negative: not allowed"),
    ],
)
def test_shape(shape, allowed):
    if allowed:
        GeoBox.from_bbox((-3, -4, 3, 4), epsg4326, shape=shape)
    else:
        with pytest.raises(ValueError):
            GeoBox.from_bbox((-3, -4, 3, 4), epsg4326, shape=shape)


@pytest.mark.parametrize(
    "gbox, expected_anchor",
    [
        (
            GeoBox.from_bbox((3, 4, 5, 7), epsg4326, resolution=0.25, anchor="edge"),
            AnchorEnum.EDGE,
        ),
        (
            GeoBox.from_bbox((-3, -4, 5, 7), epsg4326, resolution=0.25),
            AnchorEnum.EDGE,
        ),
        (
            GeoBox.from_bbox(
                (-3, -4, 5, 7), epsg4326, resolution=0.25, anchor="center"
            ),
            AnchorEnum.CENTER,
        ),
        (
            GeoBox.from_bbox((35, 40, 50, 70), epsg3857, resolution=10, tight=True),
            xy_(0.5, 0.0),
        ),
        (
            GeoBox.from_bbox((-31, 40, 50, 72), epsg3857, resolution=10, tight=True),
            xy_(0.1, 0.2),
        ),
    ],
)
def test_geobox_anchor(gbox, expected_anchor):
    assert gbox.anchor == expected_anchor



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import numpy as np
import pytest
from affine import Affine

from odc.geo import geobox as gbx
from odc.geo import wh_
from odc.geo.geobox import GeoBox
from odc.geo.testutils import epsg3857

# pylint: disable=pointless-statement,too-many-statements


def test_gbox_ops():
    s = GeoBox(wh_(1000, 100), Affine(10, 0, 12340, 0, -10, 316770), epsg3857)
    assert s.shape == (100, 1000)

    d = gbx.flipy(s)
    assert d.shape == s.shape
    assert d.crs is s.crs
    assert d.resolution.yx == (-s.resolution.y, s.resolution.x)
    assert d.extent.contains(s.extent)
    with pytest.raises(ValueError):
        # flipped grid
        (s | d)
    with pytest.raises(ValueError):
        # flipped grid
        (s & d)

    d = gbx.flipx(s)
    assert d.shape == s.shape
    assert d.crs is s.crs
    assert d.resolution.yx == (s.resolution.y, -s.resolution.x)
    assert d.extent.contains(s.extent)

    assert gbx.flipy(gbx.flipy(s)).affine == s.affine
    assert gbx.flipx(gbx.flipx(s)).affine == s.affine

    d = gbx.zoom_out(s, 2)
    assert d.shape == (50, 500)
    assert d.crs is s.crs
    assert d.extent.contains(s.extent)
    assert d.resolution.yx == (s.resolution.y * 2, s.resolution.x * 2)

    d = gbx.zoom_out(s, 2 * max(s.shape))
    assert d.shape == (1, 1)
    assert d.crs is s.crs
    assert d.extent.contains(s.extent)

    d = gbx.zoom_out(s, 1.33719)
    assert d.crs is s.crs
    assert d.extent.contains(s.extent)
    assert all(ds < ss for ds, ss in zip(d.shape, s.shape))
    with pytest.raises(ValueError):
        # lower resolution grid
        (s | d)
    with pytest.raises(ValueError):
        # lower resolution grid
        (s & d)

    d = gbx.zoom_to(s, s.shape)
    assert d == s

    d = gbx.zoom_to(s, (1, 3))
    assert d.shape == (1, 3)
    assert d.extent == s.extent

    d = gbx.zoom_to(s, (10000, 10000))
    assert d.shape == (10000, 10000)
    assert d.extent == s.extent

    d = gbx.pad(s, 1)
    assert d.crs is s.crs
    assert d.resolution == s.resolution
    assert d.extent.contains(s.extent)
    assert s.extent.contains(d.extent) is False
    assert d[1:-1, 1:-1].affine == s.affine
    assert d[1:-1, 1:-1].shape == s.shape
    assert d == (s | d)
    assert s == (s & d)

    d = gbx.pad_wh(s, 10)
    assert d == s

    d = gbx.pad_wh(s, 100, 8)
    assert d.width == s.width
    assert d.height % 8 == 0
    assert 0 < d.height - s.height < 8
    assert d.affine == s.affine
    assert d.crs is s.crs

    d = gbx.pad_wh(s, 13, 17)
    assert d.affine == s.affine
    assert d.crs is s.crs
    assert d.height % 17 == 0
    assert d.width % 13 == 0
    assert 0 < d.height - s.height < 17
    assert 0 < d.width - s.width < 13

    d = gbx.translate_pix(s, 1, 2)
    assert d.crs is s.crs
    assert d.resolution == s.resolution
    assert d.extent != s.extent
    assert s[2:3, 1:2].extent == d[:1, :1].extent

    d = gbx.translate_pix(s, -10, -2)
    assert d.crs is s.crs
    assert d.resolution == s.resolution
    assert d.extent != s.extent
    assert s[:1, :1].extent == d[2:3, 10:11].extent

    d = gbx.translate_pix(s, 0.1, 0)
    assert d.crs is s.crs
    assert d.shape == s.shape
    assert d.resolution == s.resolution
    assert d.extent != s.extent
    assert d.extent.contains(s[:, 1:].extent)

    d = gbx.translate_pix(s, 0, -0.5)
    assert d.crs is s.crs
    assert d.shape == s.shape
    assert d.resolution == s.resolution
    assert d.extent != s.extent
    assert s.extent.contains(d[1:, :].extent)

    d = gbx.affine_transform_pix(s, Affine(1, 0, 0, 0, 1, 0))
    assert d.crs is s.crs
    assert d.shape == s.shape
    assert d.resolution == s.resolution
    assert d.extent == s.extent

    d = gbx.rotate(s, 180)
    assert d.crs is s.crs
    assert d.shape == s.shape
    assert d.extent != s.extent
    np.testing.assert_almost_equal(d.extent.area, s.extent.area, 5)
    assert s[49:52, 499:502].extent.contains(
        d[50:51, 500:501].extent
    ), "Check that center pixel hasn't moved"



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Write Cloud Optimized GeoTIFFs from xarrays.
"""
from __future__ import annotations

import warnings
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple, Union
from uuid import uuid4

import numpy as np
import rasterio
import xarray as xr
from rasterio.shutil import copy as rio_copy  # pylint: disable=no-name-in-module

from ..geobox import GeoBox
from ..math import resolve_nodata
from ..types import Nodata, SomeNodata, SomeShape, shape_, wh_
from ..warp import resampling_s2rio
from ._shared import adjust_blocksize

# pylint: disable=too-many-locals,too-many-branches,too-many-arguments,too-many-statements,too-many-instance-attributes


def _without(cfg: Dict[str, Any], *skip: str) -> Dict[str, Any]:
    skip = set(skip)
    return {k: v for k, v in cfg.items() if k not in skip}


def check_write_path(fname: Union[Path, str], overwrite: bool) -> Path:
    """
    Check path before overwriting.

    :param fname: string or Path object
    :param overwrite: Whether to remove file when it exists

    exists   overwrite   Action
    ----------------------------------------------
    T            T       delete file, return Path
    T            F       raise IOError
    F            T       return Path
    F            F       return Path
    """
    if not isinstance(fname, Path):
        fname = Path(fname)

    if fname.exists():
        if overwrite:
            fname.unlink()
        else:
            raise IOError("File exists")
    return fname


def _default_cog_opts(
    *, blocksize: int = 512, shape: SomeShape = (0, 0), is_float: bool = False, **other
) -> Dict[str, Any]:
    nx, ny = shape_(shape).xy
    return {
        "tiled": True,
        "blockxsize": adjust_blocksize(blocksize, nx),
        "blockysize": adjust_blocksize(blocksize, ny),
        "zlevel": 6,
        "predictor": 3 if is_float else 2,
        "compress": "DEFLATE",
        **other,
    }


def _norm_compression_opts(
    compression: Union[bool, str, Dict[str, Any]],
    default_compress: str = "deflate",
    default_zlevel: int = 2,
) -> Dict[str, Any]:
    if isinstance(compression, bool):
        if compression:
            return {"compress": default_compress, "zlevel": default_zlevel}
        return {"compress": None}
    if isinstance(compression, str):
        compression = {"compress": compression}
    return compression


def _get_gdal_metadata(
    xx: xr.DataArray | list[xr.DataArray],
    tags: Optional[Dict[str, Any]],
) -> Dict[str, Any]:
    """
    Extract metadata from xarray for GDAL.

    :param xx: xarray.DataArray
    :return: Dictionary of metadata
    """
    meta: Dict[str, Any] = {"tags": tags}
    if not isinstance(xx, list):
        xx = [xx]

    # The logic for this is to fill the lists with
    # None values if the attribute is not present
    # as only some bands may have scale and offset
    # and we need to know which ones...
    scales, offsets, units = [], [], []
    for x in xx:
        scales.append(x.attrs.get("scales", None))
        offsets.append(x.attrs.get("offsets", None))
        units.append(x.attrs.get("units", None))

    if any(v is not None for v in scales):
        meta["scales"] = scales
    if any(v is not None for v in offsets):
        meta["offsets"] = offsets
    if any(v is not None for v in units):
        meta["units"] = units

    return meta


def _write_cog(
    pix: np.ndarray,
    geobox: GeoBox,
    fname: Union[Path, str],
    nodata: Nodata = None,
    overwrite: bool = False,
    blocksize: Optional[int] = None,
    overview_resampling: Optional[str] = None,
    overview_levels: Optional[List[int]] = None,
    ovr_blocksize: Optional[int] = None,
    use_windowed_writes: bool = False,
    intermediate_compression: Union[bool, str, Dict[str, Any]] = False,
    gdal_metadata: Optional[Dict[str, Any]] = None,
    **extra_rio_opts,
) -> Union[Path, bytes]:
    if blocksize is None:
        blocksize = 512
    if ovr_blocksize is None:
        ovr_blocksize = blocksize
    if overview_resampling is None:
        overview_resampling = "nearest"

    intermediate_compression = _norm_compression_opts(intermediate_compression)

    if pix.ndim == 2:
        h, w = pix.shape
        nbands = 1
        band = 1  # type: Any
    elif pix.ndim == 3:
        if pix.shape[:2] == geobox.shape:
            pix = pix.transpose([2, 0, 1])
        elif pix.shape[-2:] != geobox.shape:
            raise ValueError("GeoBox shape does not match image shape")

        nbands, h, w = pix.shape  # type: ignore
        band = tuple(i for i in range(1, nbands + 1))
    else:
        raise ValueError("Need 2d or 3d ndarray on input")

    assert geobox.shape == (h, w)

    if overview_levels is None:
        if min(w, h) < 512:
            overview_levels = []
        else:
            overview_levels = [2**i for i in range(1, 6)]

    if fname != ":mem:":
        path = check_write_path(
            fname, overwrite
        )  # aborts if overwrite=False and file exists already

    resampling = resampling_s2rio(overview_resampling)

    if (blocksize % 16) != 0:
        warnings.warn("Block size must be a multiple of 16, will be adjusted")

    rio_opts = {
        "width": w,
        "height": h,
        "count": nbands,
        "dtype": pix.dtype.name,
        "crs": str(geobox.crs) if geobox.crs is not None else None,
        "transform": geobox.transform,
        **_default_cog_opts(
            blocksize=blocksize, shape=wh_(w, h), is_float=pix.dtype.kind == "f"
        ),
    }
    if nodata is not None:
        rio_opts.update(nodata=nodata)

    rio_opts.update(extra_rio_opts)

    def _write(pix, band, dst):
        if not use_windowed_writes:
            dst.write(pix, band)
        else:
            for _, win in dst.block_windows():
                if pix.ndim == 2:
                    block = pix[win.toslices()]
                else:
                    block = pix[(slice(None),) + win.toslices()]

                dst.write(block, indexes=band, window=win)

        if gdal_metadata is not None:
            # Add scales, offsets and units if present
            scales = gdal_metadata.get("scales", None)
            if scales is not None:
                dst.scales = scales

            offsets = gdal_metadata.get("offsets", None)
            if offsets is not None:
                dst.offsets = offsets

            units = gdal_metadata.get("units", None)
            if units is not None:
                dst.units = units

            # Add any tags that are present
            tags = gdal_metadata.get("tags", None)
            if tags is not None:
                dst.update_tags(**tags)

    # Deal efficiently with "no overviews needed case"
    if len(overview_levels) == 0:
        if fname == ":mem:":
            with rasterio.MemoryFile() as mem:
                with mem.open(driver="GTiff", **rio_opts) as dst:
                    _write(pix, band, dst)
                return bytes(mem.getbuffer())
        else:
            with rasterio.open(path, mode="w", driver="GTiff", **rio_opts) as dst:
                _write(pix, band, dst)
            return path

    # copy re-compresses anyway so skip compression for temp image
    tmp_opts = _without(rio_opts, "compress", "predictor", "zlevel")
    tmp_opts.update(intermediate_compression)

    with rasterio.Env(GDAL_TIFF_OVR_BLOCKSIZE=ovr_blocksize):
        with rasterio.MemoryFile() as mem:
            with mem.open(driver="GTiff", **tmp_opts) as tmp:
                _write(pix, band, tmp)
                tmp.build_overviews(overview_levels, resampling)

                if fname == ":mem:":
                    with rasterio.MemoryFile() as mem2:
                        rio_copy(
                            tmp,
                            mem2.name,
                            driver="GTiff",
                            copy_src_overviews=True,
                            **_without(
                                rio_opts,
                                "width",
                                "height",
                                "count",
                                "dtype",
                                "crs",
                                "transform",
                                "nodata",
                            ),
                        )
                        return bytes(mem2.getbuffer())

                rio_copy(tmp, path, driver="GTiff", copy_src_overviews=True, **rio_opts)

    return path


def write_cog(
    geo_im: xr.DataArray,
    fname: Union[str, Path],
    *,
    overwrite: bool = False,
    blocksize: Optional[int] = None,
    ovr_blocksize: Optional[int] = None,
    overviews: Optional[Iterable[xr.DataArray]] = None,
    overview_resampling: Optional[str] = None,
    overview_levels: Optional[List[int]] = None,
    use_windowed_writes: bool = False,
    intermediate_compression: Union[bool, str, Dict[str, Any]] = False,
    tags: Optional[Dict[str, Any]] = None,
    nodata: SomeNodata = "auto",
    **extra_rio_opts,
) -> Union[Path, bytes]:
    """
    Save ``xarray.DataArray`` to a file in Cloud Optimized GeoTiff format.

    :param geo_im: ``xarray.DataArray`` with crs
    :param fname: Output path or ``":mem:"`` in which case compress to RAM and return bytes
    :param overwrite: True -- replace existing file, False -- abort with IOError exception
    :param blocksize: Size of internal tiff tiles (512x512 pixels)
    :param ovr_blocksize: Size of internal tiles in overview images (defaults to blocksize)
    :param overviews: Write pre-computed overviews if supplied
    :param overview_resampling: Use this resampling when computing overviews
    :param overview_levels: List of shrink factors to compute overiews for: [2,4,8,16,32],
                            to disable overviews supply empty list ``[]``
    :param nodata: Set ``nodata`` flag to this value if supplied, by default ``nodata`` is
                   read from the attributes of the input array (``geo_im.odc.nodata``).
    :param use_windowed_writes: Write image block by block (might need this for large images)
    :param intermediate_compression: Configure compression settings for first pass write
                    , default is no compression
    :param tags: Dictionary of tags to write into the output file. These are written as
                    GDAL Metadata items in the GeoTIFF file.
    :param extra_rio_opts: Any other option is passed to ``rasterio.open``

    :returns: Path to which output was written
    :returns: Bytes if ``fname=":mem:"``

    .. note ::

       **memory requirements**

       This function generates a temporary in memory tiff file without
       compression to speed things up. It then adds overviews to this file and
       only then copies it to the final destination with requested compression
       settings. This is necessary to produce a compliant COG, since the COG standard
       demands overviews to be placed before native resolution data and double
       pass is the only way to achieve this currently.

       This means that this function will use about 1.5 to 2 times memory taken by ``geo_im``.
    """
    nodata = resolve_nodata(nodata, geo_im.dtype, geo_im.odc.nodata)

    if overviews is not None:
        layers = [geo_im, *overviews]
        result = write_cog_layers(
            layers,
            fname,
            overwrite=overwrite,
            blocksize=blocksize,
            ovr_blocksize=ovr_blocksize,
            use_windowed_writes=use_windowed_writes,
            intermediate_compression=intermediate_compression,
            tags=tags,
            nodata=nodata,
            **extra_rio_opts,
        )
        assert result is not None
        return result

    pix = geo_im.data
    geobox = geo_im.odc.geobox

    if geobox is None:
        raise ValueError("Need geo-registered array on input")

    return _write_cog(
        pix,
        geobox,
        fname,
        nodata=nodata,
        overwrite=overwrite,
        blocksize=blocksize,
        ovr_blocksize=ovr_blocksize,
        overview_resampling=overview_resampling,
        overview_levels=overview_levels,
        use_windowed_writes=use_windowed_writes,
        intermediate_compression=intermediate_compression,
        gdal_metadata=_get_gdal_metadata(geo_im, tags),
        **extra_rio_opts,
    )


def to_cog(
    geo_im: xr.DataArray,
    blocksize: Optional[int] = None,
    ovr_blocksize: Optional[int] = None,
    overviews: Optional[Iterable[xr.DataArray]] = None,
    overview_resampling: Optional[str] = None,
    overview_levels: Optional[List[int]] = None,
    use_windowed_writes: bool = False,
    intermediate_compression: Union[bool, str, Dict[str, Any]] = False,
    tags: Optional[Dict[str, Any]] = None,
    **extra_rio_opts,
) -> bytes:
    """
    Compress ``xarray.DataArray`` into Cloud Optimized GeoTiff bytes in memory.

    This function doesn't write to disk, it compresses in RAM, which is useful
    for saving data to S3 or other cloud object stores.

    :param geo_im: ``xarray.DataArray`` with crs
    :param blocksize: Size of internal tiff tiles (512x512 pixels)
    :param ovr_blocksize: Size of internal tiles in overview images (defaults to blocksize)
    :param overviews: Write pre-computed overviews if supplied
    :param overview_resampling: Use this resampling when computing overviews
    :param overview_levels: List of shrink factors to compute overiews for: [2,4,8,16,32]
    :param nodata: Set ``nodata`` flag to this value if supplied, by default ``nodata`` is
                   read from the attributes of the input array (``geo_im.attrs['nodata']``).
    :param use_windowed_writes: Write image block by block (might need this for large images)
    :param intermediate_compression: Configure compression settings for first pass write
                    , default is no compression
    :param tags: Dictionary of tags to write into the output file. These are written as
                    GDAL Metadata items in the GeoTIFF file.
    :param extra_rio_opts: Any other option is passed to ``rasterio.open``

    :returns: In-memory GeoTiff file as bytes

    """
    bb = write_cog(
        geo_im,
        ":mem:",
        blocksize=blocksize,
        ovr_blocksize=ovr_blocksize,
        overviews=overviews,
        overview_resampling=overview_resampling,
        overview_levels=overview_levels,
        use_windowed_writes=use_windowed_writes,
        intermediate_compression=intermediate_compression,
        tags=tags,
        **extra_rio_opts,
    )

    assert isinstance(
        bb, (bytes,)
    )  # for mypy sake for :mem: output it bytes or delayed bytes
    return bb


@contextmanager
def _memfiles_ovr(nlevels) -> Generator[Tuple[rasterio.MemoryFile, ...], None, None]:
    tt = str(uuid4())
    dirname, fname = tt.split("-", 1)
    fname = fname + ".tif"
    mems = []
    try:
        for i in range(nlevels):
            mems.append(
                rasterio.MemoryFile(dirname=dirname, filename=fname + ".ovr" * i)
            )
        yield tuple(mems)
    finally:
        for mem in mems[::-1]:
            if not mem.closed:
                mem.close()
        del mems


def write_cog_layers(
    layers: Iterable[xr.DataArray],
    dst: Union[str, Path] = ":mem:",
    overwrite: bool = False,
    blocksize: Optional[int] = None,
    ovr_blocksize: Optional[int] = None,
    intermediate_compression: Union[bool, str, Dict[str, Any]] = False,
    use_windowed_writes: bool = False,
    tags: Optional[Dict[str, Any]] = None,
    nodata: Nodata = None,
    **extra_rio_opts,
) -> Union[Path, bytes, None]:
    """
    Write COG from externally computed overviews.

    Generates in-memory image with multiple side-car overviews, then re-encodes to destination with
    copy overviews flag set.
    """
    xx = list(layers)
    if len(xx) == 0:
        return None

    if dst != ":mem:":
        _ = check_write_path(dst, overwrite)

    if blocksize is None:
        blocksize = 512

    if ovr_blocksize is None:
        ovr_blocksize = blocksize

    pix = xx[0]
    gbox: GeoBox = pix.odc.geobox
    rio_opts = _default_cog_opts(
        blocksize=blocksize,
        shape=gbox.shape,
        is_float=pix.dtype.kind == "f",
        nodata=nodata,
    )
    rio_opts.update(extra_rio_opts)

    first_pass_cfg: Dict[str, Any] = {
        "num_threads": "ALL_CPUS",
        "blocksize": blocksize,
        "nodata": nodata,
        "use_windowed_writes": use_windowed_writes,
        "gdal_metadata": _get_gdal_metadata(xx, tags),
        **_norm_compression_opts(intermediate_compression),
    }

    with _memfiles_ovr(len(xx)) as mm:
        temp_fname = mm[0].name

        # write each layer into mem image
        for img, m in zip(xx, mm):
            _write_cog(
                img.data,
                img.odc.geobox,
                m.name,
                overview_levels=[],
                **first_pass_cfg,
            )

        # copy with recompression
        with rasterio.Env(
            GDAL_TIFF_OVR_BLOCKSIZE=ovr_blocksize,
            GDAL_DISABLE_READDIR_ON_OPEN=False,  # force finding of .ovr(s)
            NUM_THREADS="ALL_CPUS",
            GDAL_NUM_THREADS="ALL_CPUS",
        ):
            if dst == ":mem:":
                with rasterio.MemoryFile() as _dst:
                    rio_copy(temp_fname, _dst.name, copy_src_overviews=True, **rio_opts)
                    return bytes(_dst.getbuffer())  # makes a copy of compressed data
            else:
                rio_copy(temp_fname, dst, copy_src_overviews=True, **rio_opts)
                return Path(dst)



"""
S3 utils for COG to S3.
"""

from __future__ import annotations

from threading import Lock
from typing import TYPE_CHECKING, Any, Optional

from cachetools import cached

from ._mpu import PartsWriter, SomeData
from ._multipart import MultiPartUploadBase

if TYPE_CHECKING:
    import dask.bag
    from botocore.credentials import ReadOnlyCredentials
    from dask.delayed import Delayed

    import distributed

_state: dict[str, Any] = {}


def _mpu_local_lock(k="mpu_lock") -> Lock:
    lck = _state.get(k, None)
    if lck is not None:
        return lck

    return _state.setdefault("mpu_lock", Lock())


def _dask_client() -> "distributed.Client" | None:
    # pylint: disable=import-outside-toplevel,import-error
    from distributed import get_client

    try:
        return get_client()
    except ValueError:
        return None


def s3_parse_url(url: str) -> tuple[str, str]:
    if url.startswith("s3://"):
        bucket, *key = url[5:].split("/", 1)
        key = key[0] if len(key) else ""
        return bucket, key
    return ("", "")


class S3Limits:
    """
    Common S3 writer settings
    """

    @property
    def min_write_sz(self) -> int:
        return 5 * (1 << 20)

    @property
    def max_write_sz(self) -> int:
        return 5 * (1 << 30)

    @property
    def min_part(self) -> int:
        return 1

    @property
    def max_part(self) -> int:
        return 10_000


class S3MultiPartUpload(S3Limits, MultiPartUploadBase):
    """
    Dask to S3 dumper.
    """

    def __init__(
        self,
        bucket: str,
        key: str,
        *,
        uploadId: str = "",
        profile: Optional[str] = None,
        endpoint_url: Optional[str] = None,
        creds: Optional["ReadOnlyCredentials"] = None,
    ):
        self.bucket = bucket
        self.key = key
        self.uploadId = uploadId
        self.profile = profile
        self.endpoint_url = endpoint_url
        self.creds = creds

    @cached({})
    def s3_client(self):
        """Return the S3 client."""
        # pylint: disable=import-outside-toplevel,import-error
        from botocore.session import Session

        sess = Session(profile=self.profile)
        creds = self.creds
        if creds is None:
            return sess.create_client("s3", endpoint_url=self.endpoint_url)
        return sess.create_client(
            "s3",
            endpoint_url=self.endpoint_url,
            aws_access_key_id=creds.access_key,
            aws_secret_access_key=creds.secret_key,
            aws_session_token=creds.token,
        )

    def initiate(self, **kw) -> str:
        """Initiate the S3 multipart upload."""
        assert self.uploadId == ""
        s3 = self.s3_client()
        rr = s3.create_multipart_upload(Bucket=self.bucket, Key=self.key, **kw)
        self.uploadId = rr["UploadId"]
        return self.uploadId

    def write_part(self, part: int, data: SomeData) -> dict[str, Any]:
        """Write a single part to S3."""
        s3 = self.s3_client()
        assert self.uploadId != ""
        rr = s3.upload_part(
            PartNumber=part,
            Body=data,
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.uploadId,
        )
        return {"PartNumber": part, "ETag": rr["ETag"]}

    @property
    def url(self) -> str:
        """Return the S3 URL of the object."""
        return f"s3://{self.bucket}/{self.key}"

    def finalise(self, parts: list[dict[str, Any]]) -> str:
        """Finalise the multipart upload."""
        s3 = self.s3_client()
        assert self.uploadId
        rr = s3.complete_multipart_upload(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.uploadId,
            MultipartUpload={"Parts": parts},
        )
        return rr["ETag"]

    @property
    def started(self) -> bool:
        """Check if the multipart upload has been initiated."""
        return len(self.uploadId) > 0

    def cancel(self, other: str = ""):
        """Cancel the multipart upload."""
        uploadId = other if other else self.uploadId
        if not uploadId:
            return

        s3 = self.s3_client()
        if uploadId.lower() in ("all", ":all:"):
            for uploadId in self.list_active():
                s3.abort_multipart_upload(
                    Bucket=self.bucket, Key=self.key, UploadId=uploadId
                )
            self.uploadId = ""
        else:
            s3.abort_multipart_upload(
                Bucket=self.bucket, Key=self.key, UploadId=uploadId
            )
            if uploadId == self.uploadId:
                self.uploadId = ""

    def list_active(self) -> list[str]:
        """List active multipart uploads."""
        s3 = self.s3_client()
        rr = s3.list_multipart_uploads(Bucket=self.bucket, Prefix=self.key)
        return [x["UploadId"] for x in rr.get("Uploads", [])]

    def read(self, **kw):
        """Read the object directly from S3."""
        s3 = self.s3_client()
        return s3.get_object(Bucket=self.bucket, Key=self.key, **kw)["Body"].read()

    def __dask_tokenize__(self):
        """Dask-specific tokenization for S3 uploads."""
        return (self.bucket, self.key, self.uploadId)

    def writer(self, kw, *, client: Any = None) -> PartsWriter:
        """Return a Dask-compatible writer."""
        if client is None:
            client = _dask_client()
        writer = DelayedS3Writer(self, kw)
        if client is not None:
            writer.prep_client(client)
        return writer

    def dask_name_prefix(self) -> str:
        """Return the Dask name prefix for S3."""
        return "s3finalise"


def _safe_get(v, timeout=0.1):
    try:
        return v.get(timeout)
    except Exception:  # pylint: disable=broad-except
        return None


class DelayedS3Writer(S3Limits):
    """
    Delay multi-part upload creation until first write.
    """

    # pylint: disable=import-outside-toplevel,import-error

    def __init__(self, mpu: S3MultiPartUpload, kw: dict[str, Any]):
        self.mpu = mpu
        self.kw = kw  # mostly ContentType= kinda thing
        self._shared_var: Optional["distributed.Variable"] = None

    def prep_client(self, client: "distributed.Client") -> "distributed.Variable":
        v = self._shared(client)
        v.set(None)
        return v

    def cleanup_client(self, client: "distributed.Client") -> None:
        v = self._shared(client)
        v.delete()

    def _build_name(self, prefix: str) -> str:
        from dask.base import tokenize

        return f"{prefix}-{tokenize(self)}"

    def _shared(self, client: "distributed.Client") -> "distributed.Variable":
        from distributed import Variable

        if self._shared_var is None:
            self._shared_var = Variable(self._build_name("MPUpload"), client)
        return self._shared_var

    def _ensure_init(self, final_write: bool = False) -> S3MultiPartUpload:
        # pylint: disable=too-many-return-statements
        mpu = self.mpu
        if mpu.started:
            return mpu

        client = _dask_client()
        if client is None:
            # Assume running locally with everyone sharing same self.mpu
            with _mpu_local_lock():
                if not final_write:
                    _ = mpu.initiate(**self.kw)
                return mpu

        from distributed import Lock as DLock

        shared_state = self._shared(client)
        uploadId = _safe_get(shared_state, 0.1)

        if uploadId is not None:
            # someone else initialised it
            mpu.uploadId = uploadId
            return mpu

        lock = DLock(self._build_name("MPULock"), client)
        with lock:
            uploadId = _safe_get(shared_state, 0.1)
            if uploadId is not None:
                # someone else initialised it while we were getting a lock
                mpu.uploadId = uploadId
                return mpu

            # We are first to Lock
            # 1. Start upload
            # 2. Share UploadId with others
            if not final_write:
                _ = mpu.initiate(**self.kw)
                shared_state.set(mpu.uploadId)

        assert mpu.started or final_write
        return mpu

    def __call__(self, part: int, data: SomeData) -> dict[str, Any]:
        mpu = self._ensure_init()
        return mpu.write_part(part, data)

    def finalise(self, parts: list[dict[str, Any]]) -> Any:
        assert len(parts) > 0
        mpu = self._ensure_init()
        etag = mpu.finalise(parts)
        client = _dask_client()
        if client:
            # remove shared variable if running on distributed cluster
            self.cleanup_client(client)
        return {"Bucket": mpu.bucket, "Key": mpu.key, "ETag": etag}

    def __dask_tokenize__(self):
        return ("odc.DelayedS3Writer", self.mpu.bucket, self.mpu.key)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import array
import functools
import itertools
import math
import warnings
from collections import abc
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy
from affine import Affine
from pyproj.aoi import AreaOfInterest
from shapely import geometry, ops
from shapely.geometry import base

from ._interop import have
from .crs import CRS, CRSMismatchError, MaybeCRS, SomeCRS, norm_crs, norm_crs_or_error
from .math import edge_index, quasi_random_r2
from .types import SomeShape, SupportsCoords, Unset, shape_

CoordList = List[Tuple[float, float]]

# pylint: disable=too-many-lines,too-many-public-methods


class BoundingBox(Sequence[float]):
    """Bounding box, defining extent in cartesian coordinates."""

    __slots__ = "_crs", "_box"

    def __init__(
        self, left: float, bottom: float, right: float, top: float, crs: MaybeCRS = None
    ):
        self._box = (left, bottom, right, top)
        self._crs = norm_crs(crs)

    @property
    def left(self):
        return self._box[0]

    @property
    def bottom(self):
        return self._box[1]

    @property
    def right(self):
        return self._box[2]

    @property
    def top(self):
        return self._box[3]

    @property
    def crs(self) -> Optional[CRS]:
        return self._crs

    @property
    def bbox(self) -> Tuple[float, float, float, float]:
        return self._box

    def __iter__(self):
        return self._box.__iter__()

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, BoundingBox):
            return self._crs == other._crs and self._box == other._box
        return self._box == other

    def __hash__(self) -> int:
        return hash((self._crs, self._box))

    def __len__(self) -> int:
        return 4

    def __getitem__(self, idx):
        return self._box[idx]

    def __repr__(self) -> str:
        if self.crs is None:
            return f"BoundingBox(left={self.left}, bottom={self.bottom}, right={self.right}, top={self.top})"
        return (
            f"BoundingBox(left={self.left}, bottom={self.bottom},"
            f" right={self.right}, top={self.top}, crs={repr(self.crs)})"
        )

    def __str__(self) -> str:
        return self.__repr__()

    def __and__(self, other: "BoundingBox") -> "BoundingBox":
        return bbox_intersection([self, other])

    def __or__(self, other: "BoundingBox") -> "BoundingBox":
        return bbox_union([self, other])

    def buffered(self, xbuff: float, ybuff: Optional[float] = None) -> "BoundingBox":
        """
        Return a new BoundingBox, buffered in the x and y dimensions.

        :param xbuff: X dimension buffering amount
        :param ybuff: Y dimension buffering amount
        :return: new BoundingBox
        """
        if ybuff is None:
            ybuff = xbuff

        return BoundingBox(
            left=self.left - xbuff,
            bottom=self.bottom - ybuff,
            right=self.right + xbuff,
            top=self.top + ybuff,
            crs=self._crs,
        )

    @property
    def span_x(self) -> float:
        """Span of the bounding box along x axis."""
        return self.right - self.left

    @property
    def span_y(self) -> float:
        """Span of the bounding box along y axis."""
        return self.top - self.bottom

    @property
    def aspect(self) -> float:
        """Aspect ratio."""
        return self.span_x / self.span_y

    @property
    def width(self) -> int:
        """``int(span_x)``"""
        return int(self.right - self.left)

    @property
    def height(self) -> int:
        """``int(span_y)``"""
        return int(self.top - self.bottom)

    @property
    def shape(self) -> Tuple[int, int]:
        """``(int(span_y), int(span_x))``."""
        return (self.height, self.width)

    @property
    def range_x(self) -> Tuple[float, float]:
        """``left, right``"""
        return (self.left, self.right)

    @property
    def range_y(self) -> Tuple[float, float]:
        """``bottom, top``"""
        return (self.bottom, self.top)

    @property
    def points(self) -> CoordList:
        """Extract four corners of the bounding box."""
        x0, y0, x1, y1 = self._box
        return list(itertools.product((x0, x1), (y0, y1)))

    def transform(self, transform: Affine) -> "BoundingBox":
        """
        Map bounding box through a linear transform.

        Apply linear transform on four points of the bounding box and compute bounding box of these
        four points.
        """
        pts = [transform * pt for pt in self.points]
        xx = [x for x, _ in pts]
        yy = [y for _, y in pts]
        return BoundingBox(min(xx), min(yy), max(xx), max(yy), self._crs)

    def to_crs(self, crs: SomeCRS, **kw) -> "BoundingBox":
        """
        Compute bounding box in other CRS.
        """
        return self.polygon.to_crs(crs, **kw).boundingbox

    def map_bounds(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """
        Convert to bounds in folium/ipyleaflet style.

        Returns SW, and NE corners in lat/lon order.
        ``((lat_w, lon_s), (lat_e, lon_n))``.
        """
        x0, y0, x1, y1 = self._box
        if self._crs == "epsg:4326" or self._crs is None:
            return (y0, x0), (y1, x1)
        (x0, y0), _, (x1, y1) = self.polygon.exterior.to_crs("epsg:4326").points[:3]
        return (y0, x0), (y1, x1)

    def explore(
        self,
        map: Optional[Any] = None,
        tiles: Any = "OpenStreetMap",
        attr: Optional[str] = None,
        map_kwds: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        Plot BoundingBox on an interactive :py:mod:`folium` leaflet map
        for rapid data exploration.

        :param map:
            An optional existing :py:mod:`folium` map object to plot into.
            By default, a new map object will be created.
        :param tiles:
            Map tileset to use for the map basemap. Supports any option
            supported by :py:mod:`folium`, including "OpenStreetMap",
            "CartoDB positron", "CartoDB dark_matter" or a custom XYZ URL.
        :param attr:
            Map tile attribution; only required if passing custom tile URL.
        :param map_kwds:
            Additional keyword arguments to pass to :py:class:`folium.Map`.
        :param kwargs:
            Additional keyword arguments to pass to :py:class:`folium.GeoJson`.

        :return: A :py:mod:`folium` map containing the plotted BoundingBox.
        """
        # pylint: disable=redefined-builtin
        return self.polygon.explore(
            map=map,
            tiles=tiles,
            attr=attr,
            map_kwds=map_kwds,
            **kwargs,
        )

    @property
    def polygon(self) -> "Geometry":
        return box(*self.bbox, self.crs)

    @staticmethod
    def from_xy(
        x: Tuple[float, float], y: Tuple[float, float], crs: MaybeCRS = None
    ) -> "BoundingBox":
        """
        Construct :py:class:`~odc.geo.geom.BoundingBox` from x and y ranges.

        :param x: (left, right)
        :param y: (bottom, top)
        :param crs: CRS
        """
        x1, x2 = sorted(x)
        y1, y2 = sorted(y)
        return BoundingBox(x1, y1, x2, y2, crs)

    @staticmethod
    def from_points(
        p1: Tuple[float, float], p2: Tuple[float, float], crs: MaybeCRS = None
    ) -> "BoundingBox":
        """
        Construct :py:class:`~odc.geo.geom.BoundingBox` from two points.

        :param p1: (x, y)
        :param p2: (x, y)
        :param crs: CRS
        """
        return BoundingBox.from_xy((p1[0], p2[0]), (p1[1], p2[1]), crs)

    @staticmethod
    def from_transform(
        shape: SomeShape, transform: Affine, crs: MaybeCRS = None
    ) -> "BoundingBox":
        """
        Construct :py:class:`~odc.geo.geom.BoundingBox` from image shape and transform.

        :param shape: image dimensions
        :param transform: Affine mapping from pixel to world
        :param crs: CRS
        """
        p1 = transform * (0, 0)
        p2 = transform * shape_(shape).xy
        return BoundingBox.from_points(p1, p2, crs=crs)

    @property
    def aoi(self) -> AreaOfInterest:
        """
        Interop with pyproj.
        """
        if self._crs is None or self._crs == "epsg:4326":
            return AreaOfInterest(*self._box)
        return AreaOfInterest(*self.to_crs("epsg:4326").bbox)

    def round(self) -> "BoundingBox":
        """
        Expand bounding box to nearest integer on all sides.
        """
        x0, y0, x1, y1 = self._box
        return BoundingBox(
            math.floor(x0), math.floor(y0), math.ceil(x1), math.ceil(y1), crs=self._crs
        )

    def boundary(self, pts_per_side: int = 2) -> "Geometry":
        """
        Points along the perimeter as a linear ring.
        """
        x0, y0, x1, y1 = self._box
        xx = numpy.linspace(x0, x1, pts_per_side, dtype="float32")
        yy = numpy.linspace(y0, y1, pts_per_side, dtype="float32")
        return line(
            [
                (float(xx[ix]), float(yy[iy]))
                for iy, ix in edge_index((pts_per_side, pts_per_side), closed=True)
            ],
            self.crs,
        )

    def qr2sample(
        self,
        n: int,
        padding: Optional[float] = None,
        with_edges: bool = False,
        offset: int = 0,
    ) -> "Geometry":
        """
        Generate quasi-random sample of points within this boundingbox.

        :param n:
           Number of points

        :param padding:
           Minimal distance to the edge

        :param offset:
           Offset into quasi-random sequence from where to start

        :param edges:
           Also include samples along the edge and corners

        References: http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
        """
        # pylint: disable=too-many-locals
        x0, y0, x1, y1 = self._box
        nx = x1 - x0
        ny = y1 - y0
        pts = quasi_random_r2(n, offset=offset)
        s = numpy.asarray([nx, ny], dtype="float32")
        edge_pts: list[tuple[float, float]] = []

        if with_edges:
            sample_density = numpy.sqrt(n / (nx * ny))
            n_side = int(numpy.round(sample_density * min(nx, ny))) + 1
            n_side = max(2, n_side)
            edge_pts = [
                (float(ep[0]), float(ep[1]))
                for ep in list(self.boundary(n_side).coords[:-1])
            ]
            if padding is None:
                padding = 0.3 * min(nx, ny) / (n_side - 1)

        if padding is None:
            pts = pts * s
        else:
            pts = pts * (s - 2 * padding) + padding

        pts[:, 0] += x0
        pts[:, 1] += y0

        coords: list[tuple[float, float]] = [
            (float(p[0]), float(p[1])) for p in pts.tolist()
        ] + edge_pts

        return multipoint(coords, self.crs)


def wrap_shapely(method):
    """
    Takes a method that expects shapely geometry arguments and converts it to a method that operates
    on :py:class:`odc.geo.geom.Geometry` objects that carry their CRSs.
    """

    @functools.wraps(method, assigned=("__doc__",))
    def wrapped(*args):
        first = args[0]
        for arg in args[1:]:
            if first.crs != arg.crs:
                raise CRSMismatchError((first.crs, arg.crs))

        result = method(*[arg.geom for arg in args])
        if isinstance(result, base.BaseGeometry):
            return Geometry(result, first.crs)
        return result

    return wrapped


def force_2d(geojson: Dict[str, Any]) -> Dict[str, Any]:
    assert "type" in geojson
    assert "coordinates" in geojson

    def is_scalar(x):
        return isinstance(x, (int, float))

    def go(x):
        if is_scalar(x):
            return x

        if isinstance(x, abc.Sequence):
            if all(is_scalar(y) for y in x):
                return x[:2]
            return [go(y) for y in x]

        raise ValueError(f"invalid coordinate {x}")

    return {"type": geojson["type"], "coordinates": go(geojson["coordinates"])}


def _geojson_to_shapely(xx: Any) -> base.BaseGeometry:
    _type = xx.get("type", None)

    def to_geom(x) -> base.BaseGeometry:
        return geometry.shape(force_2d(x))

    if _type is None:
        raise ValueError("Not a valid GeoJSON")

    _type = _type.lower()
    if _type == "featurecollection":
        features = xx.get("features", [])
        if len(features) == 1:
            return to_geom(features[0]["geometry"])

        return _multigeom([to_geom(feature["geometry"]) for feature in features])

    if _type == "feature":
        return to_geom(xx["geometry"])

    return to_geom(xx)


def densify(coords: CoordList, resolution: float) -> CoordList:
    """
    Adds points so they are at most `resolution` units apart.
    """
    d2 = resolution**2

    def short_enough(p1, p2):
        return (p1[0] ** 2 + p2[0] ** 2) < d2

    new_coords = [coords[0]]
    for p1, p2 in zip(coords[:-1], coords[1:]):
        if not short_enough(p1, p2):
            segment = geometry.LineString([p1, p2])
            segment_length = segment.length
            d = resolution
            while d < segment_length:
                (pt,) = segment.interpolate(d).coords
                new_coords.append(pt)
                d += resolution

        new_coords.append(p2)

    return new_coords


def _clone_shapely_geom(geom: base.BaseGeometry) -> base.BaseGeometry:
    return geometry.shape(geom)


class Geometry(SupportsCoords[float]):
    """
    2D Geometry with CRS.

    This is a wrapper around :py:class:`shapely.geometry.BaseGeometry` that adds projection information.

    Instantiate with a GeoJSON structure. If 3D coordinates are supplied, they are converted to 2D
    by dropping the Z points.
    """

    # pylint: disable=protected-access, multiple-statements

    def __init__(
        self,
        geom: Union[base.BaseGeometry, Dict[str, Any], "Geometry"],
        crs: MaybeCRS = None,
    ):
        if isinstance(geom, Geometry):
            assert crs is None
            self.crs: Optional[CRS] = geom.crs
            self.geom: base.BaseGeometry = _clone_shapely_geom(geom.geom)
            return

        if crs is None:
            if isinstance(geom, dict) and geom.get("type", "").lower().startswith(
                "feature"
            ):
                # Assume 4326 for geojson inputs
                crs = "epsg:4326"

        crs = norm_crs(crs)
        self.crs = crs
        if isinstance(geom, base.BaseGeometry):
            self.geom = geom
        elif isinstance(geom, dict):
            self.geom = _geojson_to_shapely(geom)
        else:
            raise ValueError(f"Unexpected type {type(geom)}")

    def clone(self) -> "Geometry":
        return Geometry(self)

    # fmt: off
    @wrap_shapely
    def contains(self, other: "Geometry") -> bool: return self.contains(other)
    @wrap_shapely
    def covers(self, other: "Geometry") -> bool: return self.covers(other)
    @wrap_shapely
    def crosses(self, other: "Geometry") -> bool: return self.crosses(other)
    @wrap_shapely
    def disjoint(self, other: "Geometry") -> bool: return self.disjoint(other)
    @wrap_shapely
    def intersects(self, other: "Geometry") -> bool: return self.intersects(other)
    @wrap_shapely
    def touches(self, other: "Geometry") -> bool: return self.touches(other)
    @wrap_shapely
    def within(self, other: "Geometry") -> bool: return self.within(other)
    @wrap_shapely
    def overlaps(self, other: "Geometry") -> bool: return self.overlaps(other)
    @wrap_shapely
    def difference(self, other: "Geometry") -> "Geometry": return self.difference(other)
    @wrap_shapely
    def intersection(self, other: "Geometry") -> "Geometry": return self.intersection(other)
    @wrap_shapely
    def symmetric_difference(self, other: "Geometry") -> "Geometry": return self.symmetric_difference(other)
    @wrap_shapely
    def union(self, other: "Geometry") -> "Geometry": return self.union(other)
    @wrap_shapely
    def __and__(self, other: "Geometry") -> "Geometry": return self.__and__(other)
    @wrap_shapely
    def __or__(self, other: "Geometry") -> "Geometry": return self.__or__(other)
    @wrap_shapely
    def __xor__(self, other: "Geometry") -> "Geometry": return self.__xor__(other)
    @wrap_shapely
    def __sub__(self, other: "Geometry") -> "Geometry": return self.__sub__(other)

    def _repr_svg_(self) -> str: return self.geom._repr_svg_()

    @property
    def geom_type(self) -> str: return self.geom.geom_type
    @property
    def type(self) -> str: return self.geom.type
    @property
    def is_empty(self) -> bool: return self.geom.is_empty
    @property
    def is_valid(self) -> bool: return self.geom.is_valid
    @property
    def is_ring(self) -> bool: return self.geom.is_ring
    @property
    def boundary(self) -> "Geometry": return Geometry(self.geom.boundary, self.crs)
    @property
    def exterior(self) -> "Geometry": return Geometry(self.geom.exterior, self.crs)
    @property
    def interiors(self) -> List["Geometry"]: return [Geometry(g, self.crs) for g in self.geom.interiors]
    @property
    def centroid(self) -> "Geometry": return Geometry(self.geom.centroid, self.crs)
    @property
    def coords(self) -> CoordList: return list(self.geom.coords)
    @property
    def points(self) -> CoordList: return self.coords
    @property
    def length(self) -> float: return self.geom.length
    @property
    def area(self) -> float: return self.geom.area
    @property
    def xy(self) -> Tuple[array.array, array.array]: return self.geom.xy
    @property
    def convex_hull(self) -> "Geometry": return Geometry(self.geom.convex_hull, self.crs)
    @property
    def envelope(self) -> "Geometry": return Geometry(self.geom.envelope, self.crs)
    @property
    def boundingbox(self) -> BoundingBox:
        minx, miny, maxx, maxy = self.geom.bounds
        return BoundingBox(left=minx, right=maxx, bottom=miny, top=maxy, crs=self.crs)
    @property
    def wkt(self) -> str: return self.geom.wkt
    @property
    def __geo_interface__(self): return self.geom.__geo_interface__
    # fmt: on

    @property
    def json(self):
        return self.__geo_interface__

    def segmented(self, resolution: float) -> "Geometry":
        """
        Increase resolution of the geometry.

        Possibly add more points to the geometry so that no edge is longer than ``resolution``.
        """

        def segmentize_shapely(geom: base.BaseGeometry) -> base.BaseGeometry:
            if geom.geom_type in ["Point", "MultiPoint"]:
                return _clone_shapely_geom(geom)  # clone without changes

            if geom.geom_type in [
                "GeometryCollection",
                "MultiPolygon",
                "MultiLineString",
            ]:
                return type(geom)([segmentize_shapely(g) for g in geom.geoms])

            if geom.geom_type in ["LineString", "LinearRing"]:
                return type(geom)(densify(list(geom.coords), resolution))

            if geom.geom_type == "Polygon":
                return geometry.Polygon(
                    densify(list(geom.exterior.coords), resolution),
                    [densify(list(i.coords), resolution) for i in geom.interiors],
                )

            raise ValueError(
                f"unknown geometry type {geom.geom_type}"
            )  # pragma: no cover

        return Geometry(segmentize_shapely(self.geom), self.crs)

    def interpolate(self, distance: float) -> "Geometry":
        """
        Returns a point ``distance`` units along the line.

        @raises :py:class:`TypeError` if geometry doesn't support this operation.
        """
        return Geometry(self.geom.interpolate(distance), self.crs)

    def buffer(self, distance: float, resolution: float = 30) -> "Geometry":
        return Geometry(self.geom.buffer(distance, resolution=resolution), self.crs)

    def simplify(self, tolerance: float, preserve_topology: bool = True) -> "Geometry":
        return Geometry(
            self.geom.simplify(tolerance, preserve_topology=preserve_topology), self.crs
        )

    def transform(self, func, *, crs: MaybeCRS = Unset()) -> "Geometry":
        """
        Map through arbitrary transform.

        Applies ``func`` to all coordinates of :py:class:`~odc.geo.geom.Geometry` and returns a new
        Geometry of the same type and in the same projection from the transformed coordinates.

        ``func`` maps ``x, y``, and optionally ``z`` to output ``xp, yp, zp``. The input parameters
        may be iterable types like lists or arrays or single values. The output shall be of the same
        type: scalars in, scalars out; lists in, lists out.

        :param crs: If supplied override output CRS with that value, ``crs=None`` will remove CRS
        for the output, default is to keep the original projection.
        """
        if isinstance(func, Affine):

            def _tr(A, x, y):
                _xy = [A * xy for xy in zip(x, y)]
                return [x for x, _ in _xy], [y for _, y in _xy]

            _geom = ops.transform(functools.partial(_tr, func), self.geom)
        else:
            _geom = ops.transform(func, self.geom)

        if isinstance(crs, Unset):
            crs = self.crs

        return Geometry(_geom, crs)

    def _to_crs(self, crs: CRS) -> "Geometry":
        assert self.crs is not None
        return Geometry(ops.transform(self.crs.transformer_to_crs(crs), self.geom), crs)

    def to_crs(
        self,
        crs: SomeCRS,
        resolution: Union[float, Literal["auto"], None] = None,
        wrapdateline: bool = False,
        *,
        check_and_fix: bool = False,
    ) -> "Geometry":
        """
        Convert geometry to a different Coordinate Reference System.

        :param crs:
          CRS to convert to

        :param resolution:
          When supplied, increase resolution of the geometry before projecting.
          This is done by adding extra points such that no segment is longer
          than ``resolution`` units in the source CRS of the geometry.
          Reasonable value for the parameter depends on the application. Lower
          value will result in higher precision, but larger output geometry and
          slower processing.

          Consider using :py:meth:`~odc.geo.geom.Geometry.simplify` on the
          output of this function to reduce the size of the ouptut, especially
          when using high precision transform.

          If not supplied project the geometry as is without adding any extra points.

        :param wrapdateline:
           Attempt to gracefully handle geometry that intersects the dateline when converting to
           geographic projections. Currently only works in few specific cases (source CRS is smooth
           over the dateline).

        :param check_and_fix:
           When ``True`` remove vertices that didn't project cleanly and apply ``.buffer(0)`` is geometry
           is not valid after projection.
        """
        crs = norm_crs_or_error(crs, self)
        if self.crs == crs:
            return self

        if self.crs is None:
            raise ValueError("Cannot project geometries without CRS")

        if resolution == "auto":
            resolution = _auto_resolution(self)

        if resolution is not None and math.isfinite(resolution):
            geom = self.segmented(resolution)
        else:
            geom = self

        def maybe_fix(g: Geometry) -> Geometry:
            if not check_and_fix or g.is_valid:
                return g
            g = g.dropna()
            if g.geom_type in ["Polygon", "MultiPolygon"] and not g.is_valid:
                g = g.buffer(0)
            return g

        eps = 1e-4
        if wrapdateline and crs.geographic:
            # TODO: derive precision from resolution by converting to degrees
            precision = 0.1
            chopped = chop_along_antimeridian(geom, precision)
            chopped_lonlat = maybe_fix(chopped._to_crs(crs))
            return clip_lon180(chopped_lonlat, eps)

        return maybe_fix(geom._to_crs(crs))

    def assign_crs(self, crs: MaybeCRS) -> "Geometry":
        """
        Same geometry but with crs changed.
        """
        return Geometry(self.geom, crs=crs)

    def geojson(
        self,
        properties: Optional[Dict[str, Any]] = None,
        simplify: float = 0.05,
        resolution: Optional[float] = None,
        wrapdateline: bool = False,
        **props,
    ) -> Dict[str, Any]:
        """
        Render geometry to GeoJSON.

        Convert geometry to ``ESPG:4326`` and wrap it in GeoJSON Feature with supplied properties.

        :param properties:
            Properties to include in the GeoJSON output.

        :param simplify:
            Tolerance in degrees for simplifying geometry after changing to lon/lat. Larger number
            will result in a smaller (fewer points) and hence faster to display, but less precise
            geometry. Default is ``0.05`` of a degree. To disable set to ``0``.

        :param resolution:
           When supplied, extra points will be added to the original geometry such that no segment
           is longer than ``resolution`` units. Passed on to
           :py:meth:`~odc.geo.geom.Geometry.to_crs`.

        :param wrapdateline: Passed on to :py:meth:`~odc.geo.geom.Geometry.to_crs`

        :return: GeoJSON Feature dictionary
        """
        if self.geom_type == "GeometryCollection":
            return {
                "type": "FeatureCollection",
                "features": [
                    g.geojson(
                        properties=properties,
                        simplify=simplify,
                        resolution=resolution,
                        wrapdateline=wrapdateline,
                        **props,
                    )
                    for g in self.geoms
                ],
            }

        if self.crs is not None:
            gg = self.to_crs(
                "epsg:4326", resolution=resolution, wrapdateline=wrapdateline
            )
        else:
            gg = self

        if simplify > 0:
            gg = gg.simplify(simplify)
        if properties is None:
            properties = {**props}
        return {"type": "Feature", "geometry": gg.json, "properties": properties}

    def split(self, splitter: "Geometry") -> Iterable["Geometry"]:
        """shapely.ops.split"""
        if splitter.crs != self.crs:
            raise CRSMismatchError(self.crs, splitter.crs)

        for g in ops.split(self.geom, splitter.geom).geoms:
            yield Geometry(g, self.crs)

    def explore(
        self,
        map: Optional[Any] = None,
        tiles: Any = "OpenStreetMap",
        attr: Optional[str] = None,
        map_kwds: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Plot Geometry on an interactive :py:mod:`folium` leaflet map
        for rapid data exploration.

        :param map:
            An optional existing :py:mod:`folium` map object to plot into.
            By default, a new map object will be created.
        :param tiles:
            Map tileset to use for the map basemap. Supports any option
            supported by :py:mod:`folium`, including "OpenStreetMap",
            "CartoDB positron", "CartoDB dark_matter" or a custom XYZ URL.
        :param attr:
            Map tile attribution; only required if passing custom tile URL.
        :param map_kwds:
            Additional keyword arguments to pass to :py:class:`folium.Map`.
        :param kwargs:
            Additional keyword arguments to pass to :py:class:`folium.GeoJson`.

        :return: A :py:mod:`folium` map containing the plotted Geometry.
        """
        # pylint: disable=import-outside-toplevel, redefined-builtin

        if not have.folium:
            raise ModuleNotFoundError(
                "'folium' is required but not installed. "
                "Please install it before using `.explore()`."
            )

        from folium import Map, GeoJson

        # Create folium Map if required
        map_kwds = {} if map_kwds is None else map_kwds
        if map is None:
            map = Map(tiles=tiles, attr=attr, **map_kwds)

        # Convert to GeoJSON with resolution based on approx 100
        # points per side for proper plotting/reprojection
        bbox = self.boundingbox
        res = max(bbox.span_x, bbox.span_y) / 100
        geojson = self.geojson(resolution=res)

        # Create layer and add to map
        layer = GeoJson(data=geojson, **kwargs)
        layer.add_to(map)

        # Zoom map to extent of data
        map.fit_bounds(self.boundingbox.map_bounds())

        return map

    @property
    def geoms(self) -> Iterator["Geometry"]:
        sub_geoms = getattr(self.geom, "geoms", [])
        for geom in sub_geoms:
            yield Geometry(geom, self.crs)

    def __iter__(self) -> Iterator["Geometry"]:
        warnings.warn(
            "Iteration interface is deprecated, please use `.geoms`",
            category=DeprecationWarning,
        )
        yield from self.geoms

    def __nonzero__(self) -> bool:
        return not self.is_empty

    def __bool__(self) -> bool:
        return not self.is_empty

    def __eq__(self, other: Any) -> bool:
        return (
            hasattr(other, "crs")
            and self.crs == other.crs
            and hasattr(other, "geom")
            and self.geom == other.geom
        )

    def __str__(self):
        return f"Geometry({self.__geo_interface__}, {self.crs!r})"

    def __repr__(self):
        return f"Geometry({self.geom}, {self.crs})"

    # Implement pickle/unpickle
    # It does work without these two methods, but gdal/ogr prints 'ERROR 1: Empty geometries cannot be constructed'
    # when unpickling, which is quite unpleasant.
    def __getstate__(self):
        return {"geom": self.json, "crs": self.crs}

    def __setstate__(self, state):
        self.__init__(**state)

    @property
    def is_multi(self) -> bool:
        """True for multi-geometry types."""
        t = self.geom_type
        return t.startswith("Multi") or t == "GeometryCollection"

    def __rmul__(self, A: Affine) -> "Geometry":
        """
        Map geometry points through linear transform ``A*g``.
        """
        return self.transform(A)

    def svg(self, *args, **kw) -> str:
        """
        Returns SVG path element (wraps shapely).

        This method will drop parameters not supported by the underlying shapely geometry.

        :param scale_factor: Multiplication factor for the SVG stroke-width.  Default is 1
        :param stroke_color: Color for lines and rings
        :param fill_color: Color for poinits and polygons
        :param opacity: Float number between 0 and 1 for color opacity. Default value is 0.6
        """
        from inspect import signature  # pylint: disable=import-outside-toplevel

        valid_params = signature(self.geom.svg).parameters
        opts = {k: v for k, v in kw.items() if k in valid_params}
        return self.geom.svg(*args, **opts)

    def svg_path(self, ndecimal: int = 6) -> str:
        """
        Produce SVG path text.

        This should work for polygons/rings/lines, doesn't really
        makes sense for points.

        Example usage: ``f'<path d="{g.svg_path(0)}" ... />'``
        """

        def render_ring(g, close):
            return (
                "M"
                + "L".join(f"{x:.{ndecimal}f},{y:.{ndecimal}f}" for x, y in g.coords)
                + ("z" if close else "")
            )

        if self.is_multi:
            return "".join([g.svg_path(ndecimal) for g in self.geoms])

        if self.geom_type == "Polygon":
            return "".join(
                [render_ring(g, close=True) for g in [self.exterior, *self.interiors]]
            )

        return render_ring(self, close=self.is_ring)

    def filter(self, pred: Callable[[float, float], bool]) -> "Geometry":
        """
        Keep only those points for which `pred(x,y) is True`.
        """

        if self.geom_type == "Polygon":
            pts = [(x, y) for x, y in self.exterior.points if pred(x, y)]
            # prune inner polygon points
            inners = [
                [(x, y) for x, y in ring.points if pred(x, y)]
                for ring in self.interiors
            ]
            # prune inner polygons that don't have enough points
            inners = [ring for ring in inners if len(ring) >= 3]
            return polygon(pts, self.crs, *inners)

        if self.geom_type in ("LinearRing", "LineString"):
            pts = [(x, y) for x, y in self.points if pred(x, y)]
            if len(pts) == 1:
                pts = []  # need at least 2 points for this type
            _geom = type(self.geom)(pts)
            return Geometry(_geom, self.crs)

        if self.geom_type == "Point":
            x, y = self.coords[0]
            if pred(x, y):
                return self
            return Geometry(geometry.Point(), self.crs)

        if self.geom_type == "MultiPoint":
            pts = [(x, y) for x, y in [pt.points[0] for pt in self.geoms] if pred(x, y)]
            return multipoint(pts, self.crs)

        if self.is_multi:
            _filtered = [g.filter(pred).geom for g in self.geoms]
            _filtered = [g for g in _filtered if not g.is_empty]
            _geom = type(self.geom)(_filtered)
            return Geometry(_geom, self.crs)

        raise AssertionError("Unhandled geometry type detected")  # pragma: no cover

    def dropna(self) -> "Geometry":
        """
        Only keep finite points.
        """
        return self.filter(lambda x, y: math.isfinite(x) and math.isfinite(y))


def common_crs(geoms: Iterable[Geometry]) -> Optional[CRS]:
    """
    Compute common CRS.

    :return: CRS common across geometries.
    :return: ``None`` when input was empty

    :raises: :py:class:`odc.geo.crs.CRSMismatchError` when there are multiple.
    """
    all_crs = [g.crs for g in geoms]
    if len(all_crs) == 0:
        return None
    ref = all_crs[0]
    for crs in all_crs[1:]:
        if crs != ref:
            raise CRSMismatchError()
    return ref


def projected_lon(
    crs: MaybeCRS,
    lon: float,
    lat: Tuple[float, float] = (-90.0, 90.0),
    step: float = 1.0,
) -> Geometry:
    """
    Project vertical line along some longitude into a given CRS.

    :param crs: Destination CRS
    :param lon: Longitude to project
    :param lat: Optionally limit range of the line
    :param step: Line "resolution" in degrees
    """
    crs = norm_crs_or_error(crs, lon)
    yy = numpy.arange(lat[0], lat[1], step, dtype="float32")
    xx = numpy.full_like(yy, lon)
    tr = CRS("EPSG:4326").transformer_to_crs(crs)
    xx_, yy_ = tr(xx, yy)
    pts = [
        (float(x), float(y))
        for x, y in zip(xx_, yy_)
        if math.isfinite(x) and math.isfinite(y)
    ]
    if len(pts) < 2:
        return line([], crs)
    return line(pts, crs)


def clip_lon180(geom: Geometry, tol=1e-6) -> Geometry:
    """
    Tweak Geometry in the vicinity of longitude discontinuity.

    For every point within ``tol`` degress of ``lon=180|-180``, clip to either ``lon=180`` or
    ``lon=-180``. Which one is decided based on where the majority of other points lie.

    .. note::

      This will only do the "right thing" for "chopped" geometries. Expectation is that all the
      points are to one side of ``lon=180`` line, but some might have moved just beyond due to
      numeric tolerance issues.

    .. seealso:: :py:meth:`odc.geo.geom.chop_along_antimeridian`
    """
    thresh = 180 - tol

    def _clip_180(xx, clip):
        return [x if abs(x) < thresh else clip for x in xx]

    def _pick_clip(xx: List[float]):
        cc = 0
        for x in xx:
            if abs(x) < thresh:
                cc += 1 if x > 0 else -1
        return 180 if cc >= 0 else -180

    def transformer(xx, yy):
        clip = _pick_clip(xx)
        return _clip_180(xx, clip), yy

    if geom.geom_type.startswith("Multi"):
        return multigeom(g.transform(transformer) for g in geom.geoms)

    return geom.transform(transformer)


def chop_along_antimeridian(geom: Geometry, precision: float = 0.1) -> Geometry:
    """
    Chop a geometry along the antimeridian.

    :param geom: Geometry to maybe partition
    :param precision: in degrees
    :returns: either the same geometry if it doesn't intersect the antimeridian,
              or multi-geometry that has been split.
    """
    if geom.crs is None:
        raise ValueError("Expect geometry with CRS defined")

    l180 = projected_lon(geom.crs, 180, step=precision)
    if geom.intersects(l180):
        return multigeom(geom.split(l180))

    return geom


###########################################
# Helper constructor functions a la shapely
###########################################


def point(x: float, y: float, crs: MaybeCRS) -> Geometry:
    """
    Create a 2D Point.

    >>> point(10, 10, crs=None)
    Geometry(POINT (10 10), None)
    """
    return Geometry({"type": "Point", "coordinates": [float(x), float(y)]}, crs=crs)


def multipoint(coords: CoordList, crs: MaybeCRS) -> Geometry:
    """
    Create a 2D MultiPoint Geometry.

    >>> multipoint([(10, 10), (20, 20)], None)
    Geometry(MULTIPOINT (10 10, 20 20), None)

    :param coords: list of x,y coordinate tuples
    """
    return Geometry({"type": "MultiPoint", "coordinates": coords}, crs=crs)


def line(coords: CoordList, crs: MaybeCRS) -> Geometry:
    """
    Create a 2D LineString (Connected set of lines).

    >>> line([(10, 10), (20, 20), (30, 40)], None)
    Geometry(LINESTRING (10 10, 20 20, 30 40), None)

    :param coords: list of x,y coordinate tuples
    """
    return Geometry({"type": "LineString", "coordinates": coords}, crs=crs)


def multiline(coords: List[CoordList], crs: MaybeCRS) -> Geometry:
    """
    Create a 2D MultiLineString (Multiple disconnected sets of lines).

    >>> multiline([[(10, 10), (20, 20), (30, 40)], [(50, 60), (70, 80), (90, 99)]], None)
    Geometry(MULTILINESTRING ((10 10, 20 20, 30 40), (50 60, 70 80, 90 99)), None)

    :param coords: list of lists of x,y coordinate tuples
    """
    return Geometry({"type": "MultiLineString", "coordinates": coords}, crs=crs)


def polygon(outer, crs: MaybeCRS, *inners) -> Geometry:
    """
    Create a 2D Polygon.

    >>> polygon([(10, 10), (20, 20), (20, 10), (10, 10)], None)
    Geometry(POLYGON ((10 10, 20 20, 20 10, 10 10)), None)

    :param coords: list of 2d x,y coordinate tuples
    """
    return Geometry({"type": "Polygon", "coordinates": (outer,) + inners}, crs=crs)


def multipolygon(coords: List[List[CoordList]], crs: MaybeCRS) -> Geometry:
    """
    Create a 2D MultiPolygon.

    :param coords:
       List of polygons. Where each "polygon" is itself a list of "rings". Each "ring" is a list of
       ``(x, y)`` tuples. First ring is an outer layer of the polygon, the rest, if present, define
       holes in the outer layer.
    """
    return Geometry({"type": "MultiPolygon", "coordinates": coords}, crs=crs)


def box(
    left: float, bottom: float, right: float, top: float, crs: MaybeCRS
) -> Geometry:
    """
    Create a 2D Box (Polygon).

    >>> box(10, 10, 20, 20, None)
    Geometry(POLYGON ((10 10, 10 20, 20 20, 20 10, 10 10)), None)
    """
    points = [
        (left, bottom),
        (left, top),
        (right, top),
        (right, bottom),
        (left, bottom),
    ]
    return polygon(points, crs=crs)


def polygon_from_transform(
    shape: SomeShape, transform: Affine, crs: MaybeCRS
) -> Geometry:
    """
    Create a 2D Polygon from an affine transform and shape.

    Useful for computing footprints of a geo-registered raster images.

    :param shape: Shape of the raster in pixels
    :param transform: Affine transfrom from pixel to CRS units
    :param crs: CRS
    """
    x1, y1 = shape_(shape).xy
    points = [(0, 0), (0, y1), (x1, y1), (x1, 0), (0, 0)]
    transform.itransform(points)
    return polygon(points, crs=crs)


def sides(poly: Geometry) -> Iterable[Geometry]:
    """
    Returns a sequence of Geometry[Line] objects.

    One for each side of the exterior ring of the input polygon.
    """
    XY = poly.exterior.points
    crs = poly.crs
    for p1, p2 in zip(XY[:-1], XY[1:]):
        yield line([p1, p2], crs)


def _multigeom(geoms: List[base.BaseGeometry]) -> base.BaseGeometry:
    src_type = {g.geom_type for g in geoms}
    if len(src_type) > 1:
        return geometry.GeometryCollection(geoms)

    geoms = [g for g in geoms if not g.is_empty]
    src_type = src_type.pop()
    if src_type == "Polygon":
        return geometry.MultiPolygon(geoms)
    if src_type == "Point":
        return geometry.MultiPoint(geoms)
    if src_type == "LineString":
        return geometry.MultiLineString(geoms)
    return geometry.GeometryCollection(geoms)


def multigeom(geoms: Iterable[Geometry]) -> Geometry:
    """Construct ``Multi{Polygon|LineString|Point}``."""
    geoms = list(geoms)  # force into list
    crs = common_crs(geoms)  # will raise if some differ
    return Geometry(_multigeom([g.geom for g in geoms]), crs)


###########################################
# Multi-geometry operations
###########################################


def unary_union(geoms: Iterable[Geometry]) -> Optional[Geometry]:
    """
    Compute union of multiple (multi)polygons efficiently.
    """
    geoms = list(geoms)
    if len(geoms) == 0:
        return None

    first = geoms[0]
    crs = first.crs
    for g in geoms[1:]:
        if crs != g.crs:
            raise CRSMismatchError((crs, g.crs))

    return Geometry(ops.unary_union([g.geom for g in geoms]), crs)


def unary_intersection(geoms: Iterable[Geometry]) -> Geometry:
    """
    Compute intersection of multiple (multi)polygons.
    """
    return functools.reduce(Geometry.intersection, geoms)


def triangulate(pts: Geometry, **kw) -> Geometry:
    """
     Creates the Delaunay triangulation and returns a geometry collection.

     The source may be any geometry type. All vertices of the geometry will be
     used as the points of the triangulation.

    :param tolerance:
        From the GEOS documentation:
        tolerance is the snapping tolerance used to improve the robustness of
        the triangulation computation. A tolerance of 0.0 specifies that no
        snapping will take place.

     :param edges:
        If edges is False, a geometry collection of Polygons (triangles) will be returned (default).
        Otherwise LineString edges are returned.
    """
    geom = geometry.GeometryCollection(ops.triangulate(pts.geom, **kw))
    return Geometry(geom, pts.crs)


def intersects(a: Geometry, b: Geometry) -> bool:
    """
    Check for intersection.

    :return: ``True`` if geometries intersect, else ``False``
    """
    return a.intersects(b) and not a.touches(b)


def bbox_union(bbs: Iterable[BoundingBox]) -> BoundingBox:
    """
    Compute union of bounding boxes.

    Given a stream of bounding boxes compute enclosing :py:class:`~odc.geo.geom.BoundingBox`.
    """
    # pylint: disable=invalid-name
    try:
        bb, *bbs = bbs
    except ValueError:
        raise ValueError("Union of empty stream is undefined") from None

    L, B, R, T = bb
    crs = bb.crs

    for bb in bbs:
        l, b, r, t = bb
        L = min(l, L)
        B = min(b, B)
        R = max(r, R)
        T = max(t, T)

        if crs != bb.crs:
            raise CRSMismatchError((crs, bb.crs))

    return BoundingBox(L, B, R, T, crs)


def bbox_intersection(bbs: Iterable[BoundingBox]) -> BoundingBox:
    """
    Compute intersection of boudning boxes.

    Given a stream of bounding boxes compute the overlap :py:class:`~odc.geo.geom.BoundingBox`.
    """
    # pylint: disable=invalid-name
    try:
        bb, *bbs = bbs
    except ValueError:
        raise ValueError("Intersection of empty stream is undefined") from None

    L, B, R, T = bb
    crs = bb.crs

    for bb in bbs:
        l, b, r, t = bb
        L = max(l, L)
        B = max(b, B)
        R = min(r, R)
        T = min(t, T)

        if crs != bb.crs:
            raise CRSMismatchError((crs, bb.crs))

    return BoundingBox(L, B, R, T, crs)


def lonlat_bounds(
    geom: Geometry,
    mode: str = "safe",
    resolution: Union[float, None, Literal["auto"]] = None,
) -> BoundingBox:
    """
    Return the bounding box of a geometry in lon/lat.

    :param geom:
       Geometry in any projection
    :param mode:
       safe|quick
    :param resolution:
       If supplied will first segmentize input geometry to have no segment longer than
       ``resolution``, this increases accuracy at the cost of computation
    """
    assert mode in ("safe", "quick")
    if geom.crs is None:
        raise ValueError("lonlat_bounds can only operate on Geometry with CRS defined")

    if geom.crs.geographic:
        return geom.boundingbox

    if resolution == "auto":
        resolution = _auto_resolution(geom)

    if resolution is not None and math.isfinite(resolution):
        geom = geom.segmented(resolution)

    bbox = geom.to_crs("EPSG:4326", check_and_fix=True).boundingbox

    xx_range = bbox.range_x
    if mode == "safe":
        # If range in Longitude is more than 180 then it's probably wrapped
        # around 180 (X-360 for X > 180), so we add back 360 but only for X<0
        # values. This only works if input geometry doesn't span more than half
        # a globe, so we need to check for that too, but this is not yet
        # implemented...

        if bbox.span_x > 180:
            # TODO: check the case when input geometry spans >180 region.
            #       For now we assume "smaller" geometries not too close
            #       to poles.
            xx_ = [x + 360 if x < 0 else x for x in bbox.range_x]
            xx_range_ = min(xx_), max(xx_)
            span_x_ = xx_range_[1] - xx_range_[0]
            if span_x_ < bbox.span_x:
                xx_range = xx_range_

    return BoundingBox.from_xy(xx_range, bbox.range_y, crs=4326)


def mid_longitude(geom: Geometry) -> float:
    """
    Compute longitude of the center point of a geometry.
    """
    ((lon,), _) = geom.centroid.to_crs("epsg:4326").xy
    return lon


def _auto_resolution(g: Geometry) -> float:
    # aim for ~100 points per side of a square
    return math.sqrt(g.area) * 4 / 100



import itertools
from io import BytesIO
from pathlib import Path
from typing import Optional, Tuple

import numpy as np
import pytest

_ = pytest.importorskip("tifffile")
_ = pytest.importorskip("rasterio")

from dask import array as da
from rasterio import MemoryFile
from rasterio import open as rio_open

from odc.geo.cog import CogMeta, cog_gbox, save_cog_with_dask
from odc.geo.cog._shared import compute_cog_spec, num_overviews
from odc.geo.cog._tifffile import (
    GEOTIFF_TAGS,
    _band_names,
    _gdal_sample_description,
    _gdal_sample_descriptions,
    _make_empty_cog,
    _norm_compression_tifffile,
    _stats_from_layer,
    geotiff_metadata,
)
from odc.geo.geobox import GeoBox
from odc.geo.gridspec import GridSpec
from odc.geo.types import Unset, wh_
from odc.geo.xr import xr_zeros

gbox_globe = GridSpec.web_tiles(0)[0, 0]
_gbox = GeoBox.from_bbox((-10, -20, 15, 30), 4326, resolution=1)

# pylint: disable=redefined-outer-name


@pytest.fixture
def gbox():
    gs = GridSpec.web_tiles(2)
    return gs[2, 1]


def test_write_cog(gbox: GeoBox):
    img = xr_zeros(gbox, dtype="uint16")
    assert img.odc.geobox == gbox

    img_bytes = img.odc.to_cog(blocksize=32)
    assert isinstance(img_bytes, bytes)

    img_bytes2 = img.odc.write_cog(":mem:", blocksize=32)
    assert len(img_bytes) == len(img_bytes2)

    img_bytes = img.odc.to_cog(blocksize=32, overview_levels=[2])
    assert isinstance(img_bytes, bytes)

    img_bytes2 = img.odc.write_cog(
        ":mem:", blocksize=32, overview_levels=[2], use_windowed_writes=True
    )
    assert len(img_bytes) == len(img_bytes2)


@pytest.mark.parametrize(
    ["scales", "offsets", "units"],
    [
        (0.1, 999, "fridges"),
        (0.1, 999, None),
    ],
)
def test_write_metadata(gbox: GeoBox, scales, offsets, units):
    RANDOM = "random_value"
    img = xr_zeros(gbox, dtype="uint16")
    assert img.odc.geobox == gbox

    img.attrs.update(scales=scales, offsets=offsets, units=units)

    assert img.attrs["scales"] == scales

    img_bytes = img.odc.to_cog(
        blocksize=32,
        tags=dict(random_tag=RANDOM),
    )
    with MemoryFile(img_bytes) as memfile:
        with rio_open(memfile) as src:
            # Make sure values are set and retrieved correctly
            if isinstance(scales, (list, tuple)):
                assert src.scales == tuple(scales)
            else:
                assert src.scales[0] == scales

            if isinstance(offsets, (list, tuple)):
                assert src.offsets == tuple(offsets)
            else:
                assert src.offsets[0] == offsets

            if isinstance(units, (list, tuple)):
                assert src.units == tuple(units)
            else:
                assert src.units[0] == units

            # Check tags stick
            assert src.tags()["random_tag"] == RANDOM

            # Check the underlying data asn't altered
            assert (src.read() == img.data).all()


def test_write_cog_ovr(gbox: GeoBox):
    img = xr_zeros(gbox, dtype="uint16")
    assert img.odc.geobox == gbox
    ovrs = [img[::2, ::2], img[::4, ::4]]

    img_bytes = img.odc.to_cog(blocksize=32, overviews=ovrs)
    assert isinstance(img_bytes, bytes)

    img_bytes2 = img.odc.write_cog(":mem:", blocksize=32, overviews=ovrs)
    assert len(img_bytes) == len(img_bytes2)


@pytest.mark.parametrize(
    ["block", "dim", "n_expect"],
    [
        (2**5, 2**10, 5),
        (2**5, 2**10 - 1, 5),
        (2**5, 2**10 + 1, 5),
        (256, 78, 0),
        (1024, 1_000_000, -1),
        (512, 3_040_000, -1),
    ],
)
def test_num_overviews(block: int, dim: int, n_expect: int):
    if n_expect >= 0:
        assert num_overviews(block, dim) == n_expect
    else:
        n = num_overviews(block, dim)
        assert dim // (2**n) <= block


@pytest.mark.parametrize(
    ("shape", "tshape", "max_pad"),
    [
        [(1024, 2048), (256, 128), None],
        [(1024, 2048), (256, 128), 0],
        [(1024 + 11, 2048 + 13), (256, 128), 0],
    ],
)
def test_cog_spec(
    shape: Tuple[int, int],
    tshape: Tuple[int, int],
    max_pad: Optional[int],
):
    _shape, _tshape, nlevels = compute_cog_spec(shape, tshape, max_pad=max_pad)
    assert _shape[0] >= shape[0]
    assert _shape[1] >= shape[1]
    assert _tshape[0] >= tshape[0]
    assert _tshape[1] >= tshape[1]
    assert _tshape[0] % 16 == 0
    assert _tshape[1] % 16 == 0

    assert max(_shape) // (2**nlevels) <= max(tshape)

    if max_pad is not None:
        assert _shape[0] - shape[0] <= max_pad
        assert _shape[1] - shape[1] <= max_pad


@pytest.mark.parametrize(
    "gbox",
    [
        gbox_globe.zoom_to(shape)
        for shape in [
            (1024, 2048),
            (17, 19),
            (10_003, 10_003),
            (1 << 20, 1 << 20),
        ]
    ],
)
@pytest.mark.parametrize("kw", [{}, {"tile": 256}, {"nlevels": 5}])
def test_cog_gbox(gbox: GeoBox, kw):
    _gbox = cog_gbox(gbox, **kw)
    assert _gbox[:1, :1] == gbox[:1, :1]
    assert _gbox.shape[0] >= gbox.shape[0]
    assert _gbox.shape[1] >= gbox.shape[1]

    # once expanded has to stay same size
    assert cog_gbox(gbox, **kw) == cog_gbox(cog_gbox(gbox, **kw), **kw)


@pytest.mark.parametrize(
    "shape, blocksize, expect_ax",
    [
        ((800, 600), [400, 200], "YX"),
        ((800, 600, 3), [400, 200], "YXS"),
        ((800, 600, 4), [400, 200], "YXS"),
        ((2, 800, 600), [400, 200], "SYX"),
        ((160, 30), 16, "YX"),
        ((160, 30, 5), 16, "YXS"),
    ],
)
@pytest.mark.parametrize(
    "dtype, compression, expect_predictor",
    [
        ("int16", "deflate", 2),
        ("int16", "zstd", 2),
        ("uint8", "webp", 1),
        ("int16", Unset(), 2),
        ("float32", "zstd", 3),
        ("float32", "adobe_deflate", 3),
        ("float32", "lerc", 1),
        ("float32", "lerc_deflate", 1),
        ("float32", "lerc_zstd", 1),
    ],
)
def test_empty_cog(shape, blocksize, expect_ax, dtype, compression, expect_predictor):
    tifffile = pytest.importorskip("tifffile")
    gbox = GridSpec.web_tiles(0)[0, 0]
    if expect_ax == "SYX":
        gbox = gbox.zoom_to(shape[1:])
        assert gbox.shape == shape[1:]
    else:
        gbox = gbox.zoom_to(shape[:2])
        assert gbox.shape == shape[:2]

    meta, mm = _make_empty_cog(
        shape,
        dtype,
        gbox=gbox,
        blocksize=blocksize,
        compression=compression,
    )
    assert isinstance(mm, memoryview)
    assert meta.axis == expect_ax
    assert meta.dtype == dtype
    assert meta.shape[0] >= gbox.shape[0]
    assert meta.shape[1] >= gbox.shape[1]
    assert meta.gbox is not None
    assert meta.shape == meta.gbox.shape

    f = tifffile.TiffFile(BytesIO(mm))
    assert f.tiff.is_bigtiff

    p = f.pages[0]
    assert p.shape[0] >= shape[0]
    assert p.shape[1] >= shape[1]
    assert p.dtype == dtype
    assert p.axes == expect_ax
    assert p.predictor == expect_predictor

    if isinstance(compression, str):
        compression = compression.upper()
        if compression == "DEFLATE":
            assert p.compression.name == "ADOBE_DEFLATE"
        elif compression.startswith("LERC_"):
            assert p.compression.name == "LERC"
        else:
            assert p.compression.name == compression
    else:
        # should default to deflate
        assert p.compression == 8

    if expect_ax == "YX":
        assert f.pages[-1].chunked == (1, 1)
    elif expect_ax == "YXS":
        assert f.pages[-1].chunked[:2] == (1, 1)
    elif expect_ax == "SYX":
        assert f.pages[-1].chunked[1:] == (1, 1)

    if not isinstance(blocksize, list):
        blocksize = [blocksize]

    _blocks = itertools.chain(iter(blocksize), itertools.repeat(blocksize[-1]))
    for p, tsz in zip(f.pages, _blocks):
        if isinstance(tsz, int):
            tsz = (tsz, tsz)

        assert p.chunks[0] % 16 == 0
        assert p.chunks[1] % 16 == 0

        assert tsz[0] <= p.chunks[0] < tsz[0] + 16
        assert tsz[1] <= p.chunks[1] < tsz[1] + 16


@pytest.mark.parametrize(
    "meta",
    [
        CogMeta("YX", wh_(256, 256), wh_(128, 128), 1, "int16", 8, 1),
        CogMeta("YXS", wh_(500, 256), wh_(128, 128), 3, "uint8", 8, 1),
        CogMeta("SYX", wh_(500, 256), wh_(128, 128), 5, "float32", 8, 1),
    ],
)
def test_cog_meta(meta: CogMeta):
    for idx_flat, idx in enumerate(meta.tidx()):
        assert meta.flat_tile_idx(idx) == idx_flat

    assert meta.num_tiles == (meta.chunked.x * meta.chunked.y * meta.num_planes)
    assert meta.num_tiles == len(list(meta.tidx()))

    assert len(meta.pix_shape) == {"YX": 2, "SYX": 3, "YXS": 3}[meta.axis]
    assert len(meta.pix_shape) == len(meta.chunks)

    if meta.axis in ("YX", "YXS"):
        assert meta.num_planes == 1
        assert meta.nsamples >= meta.num_planes

    if meta.axis == "SYX":
        assert meta.num_planes == meta.nsamples

    layers = meta.flatten()
    for idx in meta.cog_tidx():
        assert isinstance(idx, tuple)
        assert len(idx) == 4
        img_idx, s, y, x = idx
        tidx = (s, y, x)
        assert 0 <= img_idx < len(layers)
        assert layers[img_idx].flat_tile_idx(tidx) <= layers[img_idx].num_tiles

    for bad_idx in [
        (0, -1, 0),
        (-1, 0, 0),
        (0, 0, -1),
        (meta.num_planes, 0, 0),
        (0, meta.chunked.y, 0),
        (0, meta.chunked.y - 1, meta.chunked.x + 2),
    ]:
        with pytest.raises(IndexError):
            _ = meta.flat_tile_idx(bad_idx)


def test_norm_compress():
    predictor, compression, ca = _norm_compression_tifffile("int16")
    assert compression == "ADOBE_DEFLATE"
    assert predictor == 2
    assert ca == {}


@pytest.mark.parametrize(
    "gbox",
    [
        _gbox,
        _gbox.to_crs(3857),
        _gbox.to_crs("ESRI:53010"),
        _gbox.rotate(10),
        _gbox.center_pixel.pad(3),
    ],
)
@pytest.mark.parametrize("nodata", ["auto", None, float("nan"), 0, -999])
@pytest.mark.parametrize("gdal_metadata", [None, "<GDALMetadata></GDALMetadata>"])
def test_geotiff_metadata(gbox: GeoBox, nodata, gdal_metadata: Optional[str]):
    assert gbox.crs is not None

    geo_tags, md = geotiff_metadata(gbox, nodata=nodata, gdal_metadata=gdal_metadata)
    assert isinstance(md, dict)
    assert isinstance(geo_tags, list)
    assert len(geo_tags) >= 2
    tag_codes = {code for code, *_ in geo_tags}

    if nodata is not None:
        assert 42113 in tag_codes
    if gdal_metadata is not None:
        assert 42112 in tag_codes

    for code, dtype, count, val in geo_tags:
        assert code in GEOTIFF_TAGS
        assert isinstance(dtype, int)
        assert isinstance(count, int)
        if count > 0:
            assert isinstance(val, (tuple, str))
            if isinstance(val, str):
                assert len(val) + 1 == count
            else:
                assert len(val) == count

    if gbox.axis_aligned:
        assert "ModelPixelScale" in md
    else:
        assert "ModelTransformation" in md

    if gbox.crs.epsg is not None:
        if gbox.crs.projected:
            assert md["GTModelTypeGeoKey"] == 1
            assert md["ProjectedCSTypeGeoKey"] == gbox.crs.epsg
        else:
            assert md["GTModelTypeGeoKey"] == 2
            assert md["GeographicTypeGeoKey"] == gbox.crs.epsg
    else:
        assert md["GTModelTypeGeoKey"] == 32767


@pytest.mark.parametrize(
    ("sample", "description", "expected"),
    [
        (
            0,
            "easy",
            '<Item name="DESCRIPTION" sample="0" role="description">easy</Item>',
        ),
        (
            1,
            "<",
            '<Item name="DESCRIPTION" sample="1" role="description">&amp;lt;</Item>',
        ),
        (
            2,
            ">",
            '<Item name="DESCRIPTION" sample="2" role="description">&amp;gt;</Item>',
        ),
        (
            3,
            "&",
            '<Item name="DESCRIPTION" sample="3" role="description">&amp;amp;</Item>',
        ),
        (
            4,
            "& <a>",
            '<Item name="DESCRIPTION" sample="4" role="description">&amp;amp; &amp;lt;a&amp;gt;</Item>',
        ),
    ],
)
def test_gdal_sample_description(sample: int, description: str, expected: str):
    assert _gdal_sample_description(sample, description) == expected


def test_gdal_sample_descriptions():
    assert _gdal_sample_descriptions(["red", "green", "blue"]) == [
        '<Item name="DESCRIPTION" sample="0" role="description">red</Item>',
        '<Item name="DESCRIPTION" sample="1" role="description">green</Item>',
        '<Item name="DESCRIPTION" sample="2" role="description">blue</Item>',
    ]


def test_band_names(gbox: GeoBox):
    gbox = gbox.zoom_to(1024)
    dtype = "float32"
    n = 512
    nodata = -9999

    img = xr_zeros(gbox, dtype, chunks=(n, n), nodata=nodata)
    img.attrs["long_name"] = []

    assert _band_names(img) == []

    img.attrs["long_name"] = "first band"
    assert _band_names(img) == ["first band"]

    img.attrs["long_name"] = ["first band"]
    assert _band_names(img) == ["first band"]

    img = img.odc.colorize()
    assert img.ndim == 3

    # Obtain descriptions from strings in "band" coordinate
    assert _band_names(img) == ["r", "g", "b", "a"]

    # Obtain descriptions from "long_names" attribute
    img["band"] = [0, 1, 2, 3]
    img.attrs["long_name"] = ["red", "green", "blue", "alpha"]
    assert _band_names(img) == ["red", "green", "blue", "alpha"]


@pytest.mark.parametrize("dtype", ["int16", "float32"])
def test_cog_with_dask_smoke_test(gbox: GeoBox, tmp_path: Path, dtype):
    gbox = gbox.zoom_to(1024)
    assert gbox.shape == (1024, 1024)
    n = 512
    nodata = -9999

    fname = str(tmp_path / "cog.tif")
    img = xr_zeros(gbox, dtype, chunks=(n, n), nodata=nodata)
    fut = save_cog_with_dask(img, fname, compression="deflate", level=2)
    assert fut is not None
    rr = fut.compute()
    assert str(rr) == fname

    # YXS
    img = img.odc.colorize()
    assert img.ndim == 3
    assert img.odc.ydim == 0

    fname = str(tmp_path / "cog-rgba.tif")
    fut = save_cog_with_dask(img, fname, compression="deflate", level=2)
    rr = fut.compute()
    assert str(rr) == fname

    # SYX, one band
    img = xr_zeros(gbox, dtype, chunks=(1, n, n), time=["2000"])
    assert img.ndim == 3
    assert img.odc.ydim == 1
    fname = str(tmp_path / "cog-syx-one-band.tif")
    fut = save_cog_with_dask(img, fname, compression="deflate", level=2)
    rr = fut.compute()
    assert str(rr) == fname

    # SYX, multiple bands
    img = xr_zeros(gbox, dtype, chunks=(1, n, n), time=["2000", "2001"])
    assert img.ndim == 3
    assert img.odc.ydim == 1
    fname = str(tmp_path / "cog-syx.tif")
    fut = save_cog_with_dask(img, fname, compression="deflate", level=2)
    rr = fut.compute()
    assert str(rr) == fname

    # Band names
    img.attrs["long_name"] = ["2000", "2001"]
    fname = str(tmp_path / "cog-bandnames.tif")
    fut = save_cog_with_dask(img, fname, compression="deflate", level=2)
    rr = fut.compute()
    assert str(rr) == fname

    # Incorrect number of band names
    img.attrs["long_name"] = ["red", "green", "blue"]
    fname = str(tmp_path / "cog-bandnames-incorrect.tif")
    with pytest.raises(ValueError):
        save_cog_with_dask(img, fname, compression="deflate", level=2)


@pytest.mark.parametrize(
    ("array", "nodata", "minimum", "maximum", "mean", "stddev", "valid_percent"),
    [
        pytest.param([[1, 1], [1, 1]], None, 1, 1, 1, 0, 100, id="basic int"),
        pytest.param(
            [[1.0, 1.0], [1.0, 1.0]], None, 1.0, 1.0, 1.0, 0.0, 100, id="basic float"
        ),
        pytest.param([[1, 0], [0, 1]], 0, 1, 1, 1, 0, 50, id="int with numeric nodata"),
        pytest.param(
            [[1.0, 0.0], [0.0, 1.0]], 0, 1, 1, 1, 0, 50, id="float with numeric nodata"
        ),
        pytest.param(
            [[1.0, np.nan], [np.nan, 1.0]],
            None,
            1,
            1,
            1,
            0,
            50,
            id="float with nan, None nodata",
        ),
        pytest.param(
            [[1.0, np.nan], [np.nan, 1.0]],
            np.nan,
            1,
            1,
            1,
            0,
            50,
            id="float with nan, nan nodata",
        ),
        pytest.param(
            [
                [1.0, np.nan],
                [0.0, 9.0],
            ],
            0,
            1,
            9.0,
            5.0,
            4.0,
            50,
            id="float with nan, numeric nodata",
        ),
    ],
)
def test_stats_from_layer(array, nodata, minimum, maximum, mean, stddev, valid_percent):
    x = da.from_array(array)
    stats = _stats_from_layer(x, nodata).compute()[0]

    assert stats["minimum"] == minimum
    assert stats["maximum"] == maximum
    assert stats["mean"] == mean
    assert stats["stddev"] == stddev
    assert stats["valid_percent"] == valid_percent
    assert stats["valid_percent"] == valid_percent



import itertools
import math
from functools import lru_cache
from textwrap import dedent
from typing import Any, Optional, Tuple

from . import geom
from .crs import CRS
from .types import XY, OutlineMode, Shape2d, wh_, xy_

# pylint: disable=too-many-locals,import-outside-toplevel
_UNIT_REMAPS = {
    "metre": "m",
    "meter": "m",
    "degrees": "°",
    "degrees_north": "°",
    "degrees_south": "°",
    "degrees_east": "°",
    "degrees_west": "°",
}


def norm_units(unit: str) -> str:
    return _UNIT_REMAPS.get(unit, unit)


def pick_grid_step(N: int, at_least: int = 4, no_more: int = 11) -> int:
    if N <= 0:
        return 1

    factors = [1, 5, 10, 2, 4, 3, 6, 7, 8, 9, 1.5, 2.5]
    n = 10 ** (math.floor(math.log10(N)) - 1)
    if n < 1:
        return 1

    for f in factors:
        step = int(n * f)
        count = N // step
        if at_least <= count < no_more:
            return step

    return N // at_least


@lru_cache
def _ocean_svg_path(ndecimal=3, clip_bbox=None):
    from .data import ocean_geom

    g = ocean_geom(bbox=clip_bbox)
    return g.svg_path(ndecimal)


def _compute_display_box(
    span: XY[float], sz: int, min_sz: int, tol: float = 1e-6
) -> Tuple[Shape2d, float]:
    """
    return: shape in pixels, ``s`` maps sizes from world to pix
    """
    span_x, span_y = ((v if math.isfinite(v) else 0) for v in span.xy)
    if max(span_x, span_y) < tol:
        # both too small, make it square
        span_x = span_y = tol
        w, h = sz, sz
        s = w / span_x
    elif span_x > span_y:
        # wide case: w = sz, span_x == sz
        w, h = sz, max(int(sz * span_y / span_x), min_sz)
        s = w / span_x
    else:
        # tall case: h = sz, span_y == sz
        h, w = sz, max(int(sz * span_x / span_y), min_sz)
        s = h / span_y
    return wh_(w, h), s


def svg_base_map(
    *extras,
    fill="#CCCCCC",
    opacity=1,
    stroke_width=1,
    stroke="#555555",
    stroke_opacity=0.9,
    bbox=None,
    sz=360,
    map_svg_path=None,
    target=None,
):
    clip_bbox = None
    if bbox is not None:
        clip_bbox = geom.bbox_intersection(
            [geom.BoundingBox(*bbox).buffered(1), geom.BoundingBox(-180, -90, 180, 90)]
        ).bbox
    else:
        bbox = (-180, -90, 180, 90)

    if map_svg_path is None:
        map_svg_path = _ocean_svg_path(clip_bbox=clip_bbox)

    x0, y0, x1, y1 = bbox
    shape, s = _compute_display_box(xy_(x1 - x0, y1 - y0), sz, min(40, sz))
    w, h = shape.wh

    scale_factor = 1 / s
    stroke_width = stroke_width * scale_factor

    more_svg = ""
    for x in extras:
        if isinstance(x, str):
            more_svg += x
        else:
            more_svg += x.svg(scale_factor)

    if target is not None:
        target_x, target_y = target
        more_svg += (
            f'<path stroke="{stroke}" stroke-width="{stroke_width}" stroke-opacity="{stroke_opacity*0.8}"'
            f' d="M{target_x},{target_y} V90zV-90zH180zH-180z"/>'
        )

    return f"""\
<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg"
     xmlns:xlink="http://www.w3.org/1999/xlink"
     width="{w:d}" height="{h:d}"
     viewBox="0 0 {w:d} {h:d}"
     preserveAspectRatio="xMinYMin meet">
<g transform="matrix({s},0,0,{-s},{-x0*s},{y1*s})">
<path fill-rule="evenodd"
  fill="{fill}"
  opacity="{opacity}"
  stroke="{stroke}"
  stroke-width="{stroke_width}"
  stroke-opacity="{stroke_opacity}"
  d="{map_svg_path}"/>
{more_svg}</g></svg>"""


def make_svg(
    *extras,
    stroke_width=1,
    bbox=None,
    sz=360,
):
    if bbox is None:
        bbox = (0, 0, sz, sz)

    x0, y0, x1, y1 = bbox
    shape, s = _compute_display_box(xy_(x1 - x0, y1 - y0), sz, min(40, sz))
    w, h = shape.wh

    scale_factor = 1 / s
    stroke_width = stroke_width * scale_factor

    more_svg = ""
    for x in extras:
        if isinstance(x, str):
            more_svg += x
        else:
            more_svg += x.svg(scale_factor)

    return f"""\
<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg"
     xmlns:xlink="http://www.w3.org/1999/xlink"
     width="{w:d}" height="{h:d}"
     viewBox="0 0 {w:d} {h:d}"
     preserveAspectRatio="xMinYMin meet">
<g transform="matrix({s},0,0,{-s},{-x0*s},{y1*s})">
{more_svg}
</g></svg>"""


class PixelGridDisplay:
    """
    Common code for visualization of geoboxes.

    """

    def __init__(
        self,
        src: Any,
        pix2world: Any,
        gsd: float,
    ) -> None:
        self._src = src
        self._shape: Shape2d = src.shape
        self._crs: Optional[CRS] = src.crs
        self._pix2world = pix2world
        self._gsd = gsd

    def svg(
        self,
        scale_factor: float = 1.0,
        mode: OutlineMode = "auto",
        notch: float = 0.0,
        grid_stroke: str = "pink",
    ) -> str:
        """
        Produce SVG paths.

        :param mode: One of pixel, native, geo (default is geo)
        :return: SVG path
        """
        if mode == "auto":
            mode = "native" if self._crs is None else "geo"

        grids = self.grid_lines(mode=mode)
        outline = self.outline(mode, notch=notch)

        grid_svg = (
            '<path fill="none" opacity="0.8"'
            f' stroke-width="{0.8*scale_factor}"'
            f' stroke="{grid_stroke}"'
            f' d="{grids.svg_path()}" />'
        )

        return outline.svg(scale_factor) + grid_svg

    def grid_lines(self, step: int = 0, mode: OutlineMode = "native") -> geom.Geometry:
        """
        Construct pixel edge aligned grid lines.
        """

        nx, ny = self._shape.xy
        if nx > 0 and ny > 0:
            if step == 0:
                step = pick_grid_step(max(nx, ny))
            xx = [*range(0, nx, step), nx]
            yy = [*range(0, ny, step), ny]
            vertical = [list(itertools.product([x], yy)) for x in xx[1:-1]]
            horizontal = [list(itertools.product(xx, [y])) for y in yy[1:-1]]
            lines = geom.multiline(vertical + horizontal, self._crs)
        else:
            lines = geom.multiline([], self._crs)

        if mode == "pixel":
            return lines

        lines = lines.transform(self._pix2world)
        if mode == "native":
            return lines

        dx, dy = self._pix2world(step, 0)
        res = math.sqrt(dx * dx + dy * dy) / 5
        return lines.to_crs("epsg:4326", resolution=res).dropna()

    def outline(
        self, mode: OutlineMode = "native", notch: float = 0.1
    ) -> geom.Geometry:
        """
        Produce Line Geometry around perimeter.

        .. code-block:: txt

             +---+-------------+
             |   |             |
             +---+             |
             |                 |
             |                 |
             +-----------------+
        """

        assert notch < 1
        w, h = self._shape.wh
        if notch > 0:
            nn = min(notch * max(w, h), w, h)
            pix = geom.line(
                [
                    (0, nn),
                    (0, 0),
                    (nn, 0),
                    (w, 0),
                    (w, h),
                    (0, h),
                    (0, nn),
                    (nn, nn),
                    (nn, 0),
                ],
                self._crs,
            )
        else:
            pix = geom.multigeom(
                [
                    geom.line([(0, 0), (w, 0), (w, h), (0, h), (0, 0)], self._crs),
                    geom.point(0, 0, self._crs),
                ]
            )
        if mode == "pixel":
            return geom.Geometry(pix.geom, crs=None)

        native = pix.transform(self._pix2world)
        if mode == "native":
            return native

        # about 100 pts per side
        bbox = native.boundingbox
        res = max(bbox.span_x, bbox.span_y) / 100

        return native.to_crs("EPSG:4326", resolution=res).dropna()

    def _display_bbox(self, pad_fraction: float = 0.1):
        bbox = self._src.geographic_extent.boundingbox
        pad_deg = max(bbox.span_x, bbox.span_y) * pad_fraction
        return bbox.buffered(pad_deg)

    def _render_svg(self, sz=360):
        if self._crs is None:
            bbox = self._src.extent.boundingbox
            margin = 0.1 * max(bbox.span_x, bbox.span_y)
            bbox = bbox.buffered(margin)
            return make_svg(
                self,
                bbox=bbox,
                sz=sz,
            )

        return svg_base_map(self, bbox=self._display_bbox(), sz=sz)

    def _repr_svg_(self):
        return self._render_svg()

    def _repr_html_(self):
        from .data import gbox_css

        W, H = self._shape.wh
        grid_step = pick_grid_step(max(W, H))
        svg_zoomed_txt = self._render_svg(sz=320)

        crs = self._crs
        if crs is None:
            authority = ("", "")
            wkt = "not set"
            units = ""
            svg_global_txt = ""
        else:
            authority = crs.authority
            wkt = crs.to_wkt(pretty=True).replace("\n", "<br/>").replace(" ", "&nbsp;")
            units = crs.units[0]
            svg_global_txt = svg_base_map(
                sz=200, target=self._src.geographic_extent.centroid.coords[0]
            )

        if authority == ("", ""):
            authority = ("CRS", "WKT")

        units = norm_units(units)
        pix_sz = self._gsd

        info = [
            ("Dimensions", f"{W:,d}x{H:,d}"),
            authority,
            ("Resolution", f"{pix_sz:g}{units}"),
            ("Cell", f"{grid_step:,d}px"),
        ]

        info_html = "\n".join(
            [
                (
                    f'<div class="row"><div class="column">{hdr}</div>'
                    f'<div class="column value">{val}</div></div>'
                )
                for hdr, val in info
            ]
        )
        src_class_name = type(self._src).__name__

        return dedent(
            f"""\
        <style>{gbox_css()}</style>
        <div class="gbox-info">
        <h4>{src_class_name}</h4>
        <div class="row">
            <div class="column">
                <div class="info-box">
                    {info_html}
                    <div>{svg_global_txt}</div>
                </div>
            </div>
            <div class="column svg-zoomed">{svg_zoomed_txt}</div>
        </div>
        <details>
            <summary>WKT</summary>
            <div class="wkt">{wkt}</div>
        </details>
        </div>"""
        )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import itertools
import math

import numpy
import pytest
from pytest import approx

from odc.geo import CRS, res_, resyx_, xy_, yx_
from odc.geo.geom import polygon
from odc.geo.gridspec import GridSpec
from odc.geo.testutils import SAMPLE_WKT_WITHOUT_AUTHORITY

# pylint: disable=protected-access,use-implicit-booleaness-not-comparison
# pylint: disable=comparison-with-itself,unnecessary-comprehension


def test_gridspec():
    gs = GridSpec(
        crs=CRS("EPSG:4326"),
        tile_shape=(10, 10),
        resolution=0.1,
        origin=xy_(10, 10),
    )
    assert gs.tile_shape == (10, 10)
    assert gs.tile_size == xy_(1, 1)

    assert gs == GridSpec(
        gs.crs, gs.tile_shape, resolution=resyx_(-0.1, 0.1), origin=gs.origin
    )
    assert gs == GridSpec(gs.crs, gs.tile_shape, resolution=res_(0.1), origin=gs.origin)

    poly = polygon(
        [(10, 12.2), (10.8, 13), (13, 10.8), (12.2, 10), (10, 12.2)],
        crs=CRS("EPSG:4326"),
    )
    cells = {index: geobox for index, geobox in list(gs.tiles_from_geopolygon(poly))}
    assert set(cells.keys()) == {(0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1)}
    assert numpy.isclose(
        cells[(2, 0)].coordinates["longitude"].values,
        numpy.linspace(12.05, 12.95, num=10),
    ).all()
    assert numpy.isclose(
        cells[(2, 0)].coordinates["latitude"].values,
        numpy.linspace(10.95, 10.05, num=10),
    ).all()

    # check geobox_cache
    cache = {}
    poly = gs[3, 4].extent
    ((c1, gbox1),) = list(gs.tiles_from_geopolygon(poly, geobox_cache=cache))
    ((c2, gbox2),) = list(gs.tiles_from_geopolygon(poly, geobox_cache=cache))

    assert c1 == (3, 4) and c2 == c1
    assert gbox1 is gbox2

    assert "4326" in str(gs)
    assert "4326" in repr(gs)
    assert gs == gs
    assert (gs == {}) is False
    assert gs.dimensions == ("latitude", "longitude")

    assert GridSpec("epsg:3857", (100, 100), 1).alignment == xy_(0, 0)
    assert GridSpec("epsg:3857", (100, 100), resyx_(1, 1)).dimensions == ("y", "x")

    assert GridSpec("epsg:3857", (10, 20), 11.0) == GridSpec(
        "epsg:3857", (10, 20), resyx_(-11, 11)
    )
    assert GridSpec("epsg:3857", (10, 20), 11) == GridSpec(
        "epsg:3857", (10, 20), resyx_(-11, 11)
    )

    # missing shape parameter
    with pytest.raises(ValueError):
        GridSpec.from_sample_tile(poly)


def test_non_square():
    gs = GridSpec(
        crs=CRS("EPSG:4326"),
        tile_shape=(30, 10),
        resolution=0.1,
        origin=xy_(0.05, 0.0),
    )
    assert gs.tile_shape == (30, 10)
    assert gs.tile_size == yx_(3, 1)
    assert gs.alignment == xy_(0.05, 0)


def test_web_tiles():
    TSZ0 = 6_378_137 * 2 * math.pi
    epsg3857 = CRS("epsg:3857")

    gs = GridSpec.web_tiles(0)
    assert gs.crs == epsg3857
    assert gs.tile_shape == (256, 256)
    assert gs.tile_size.xy == approx((TSZ0, TSZ0))

    gs = GridSpec.web_tiles(1)
    assert gs.crs == epsg3857
    assert gs.tile_shape == (256, 256)
    assert gs.tile_size.xy == approx((TSZ0 / 2, TSZ0 / 2))


@pytest.mark.parametrize(
    "gs, tiles_to_check",
    [
        (GridSpec.web_tiles(3), itertools.product(range(8), range(8))),
        (GridSpec(4326, (360, 360), 0.01), [(0, 0), (1, 3), (2, 1)]),
    ],
)
def test_tiles_tight_query_issue_97(gs: GridSpec, tiles_to_check):
    for idx in tiles_to_check:
        bbox = gs[idx].boundingbox
        tiles = list(gs.tiles(bbox))
        assert len(tiles) == 1
        ((_idx, gbox),) = tiles
        assert idx == _idx
        assert gs[idx] == gbox


def test_geojson():
    gs = GridSpec.web_tiles(3)
    gjson = gs.geojson()

    assert gjson["type"] == "FeatureCollection"
    assert len(gjson["features"]) == (2**3) ** 2

    gjson = gs.geojson(geopolygon=gs.crs.valid_region.buffer(-0.1))
    assert len(gjson["features"]) == (2**3) ** 2

    gjson = gs.geojson(
        bbox=gs.crs.valid_region.buffer(-0.1).to_crs("epsg:3857").boundingbox
    )
    assert len(gjson["features"]) == (2**3) ** 2

    crs = CRS(SAMPLE_WKT_WITHOUT_AUTHORITY)
    gs = GridSpec(crs, (10, 10), resolution=6_378_137 * 2 * math.pi)
    gjson = gs.geojson()
    assert len(gjson["features"]) > 0



"""
MPU file sink
"""

from __future__ import annotations

import mmap
from pathlib import Path
from typing import Any

from ._mpu import SomeData


class MPUFileSink:
    """
    Writes chunks to disk, then gathers them into final destination.

    Assumes shared filesystem across Dask cluster.
    """

    def __init__(
        self,
        dst: str | Path,
        parts_base: str | Path | None = None,
        **limits,
    ) -> None:
        dst = Path(dst)
        if parts_base is None:
            parts_dir = dst.parent / f".{dst.name}.parts"
        else:
            parts_dir = Path(parts_base) / f".{dst.name}.parts"

        self._dst = dst
        self._parts_dir = parts_dir
        self._limits = limits

    @property
    def min_write_sz(self) -> int:
        return self._limits.get("min_write_sz", 4096)

    @property
    def max_write_sz(self) -> int:
        return self._limits.get("min_write_sz", 5 * (1 << 30))

    @property
    def min_part(self) -> int:
        return self._limits.get("min_part", 1)

    @property
    def max_part(self) -> int:
        return self._limits.get("min_part", 10_000)

    def _ensure_dst_file(self, part: int) -> Path:
        parts_dir = self._parts_dir
        if not parts_dir.exists():
            try:
                parts_dir.mkdir(parents=True)
            except FileExistsError:
                pass
        return parts_dir / f"p{part:04d}.bin"

    def __call__(self, part: int, data: SomeData) -> dict[str, Any]:
        dst = self._ensure_dst_file(part)
        with open(dst, "wb") as f:
            nb = f.write(data)
            assert nb == len(data)

        return {"PartNumber": part, "Path": str(dst), "Size": len(data)}

    def finalise(self, parts: list[dict[str, Any]], keep_parts: bool = False) -> Any:
        assert len(parts) > 0
        dst = self._dst
        first, *rest = parts
        p1 = Path(first["Path"])
        p1.rename(dst)

        with open(dst, "ab") as f:
            for part in rest:
                src_path = Path(part["Path"])
                with src_path.open("rb") as src:
                    with mmap.mmap(
                        src.fileno(), 0, access=mmap.ACCESS_READ
                    ) as src_bytes:
                        f.write(src_bytes)

                if not keep_parts:
                    src_path.unlink()

        if not keep_parts:
            self._parts_dir.rmdir()

        return dst

    def __dask_tokenize__(self):
        return (self._dst, self._parts_dir)
